import uuid

from flask import Blueprint, Response, jsonify, request

from hiveos.api.context import get_core_from_request
from hiveos import config as config_module
from hiveos.intelligence.model_routing_artifacts import (
    create_model_routing_artifact,
    get_model_routing_artifact,
    list_model_routing_artifacts,
)
from hiveos.intelligence.model_routing import build_model_routing_recommendation
from hiveos.intelligence.tracing import (
    build_trace_event,
    emit_trace,
    read_trace_events,
    read_trace_events_ndjson,
)


traces_bp = Blueprint("traces", __name__)


@traces_bp.route("/trace_events", methods=["GET"])
def trace_events():
    try:
        _ = get_core_from_request()
    except PermissionError as e:
        return jsonify({"error": str(e)}), 403

    limit = request.args.get("limit", default=200, type=int)
    if limit > 1000:
        limit = 1000

    filters = {
        "run_id": request.args.get("run_id"),
        "agent_id": request.args.get("agent_id"),
        "event_type": request.args.get("event_type"),
        "outcome": request.args.get("outcome"),
        "task_id": request.args.get("task_id"),
        "chunk_id": request.args.get("chunk_id"),
        "zone_id": request.args.get("zone_id"),
        "model_provider": request.args.get("model_provider"),
        "model_name": request.args.get("model_name"),
        "quality_status": request.args.get("quality_status"),
    }
    events = read_trace_events(limit=limit, filters=filters)
    return jsonify({"events": events, "count": len(events), "limit": limit})


@traces_bp.route("/trace_events_export", methods=["GET"])
def trace_events_export():
    try:
        _ = get_core_from_request()
    except PermissionError as e:
        return jsonify({"error": str(e)}), 403

    limit = request.args.get("limit", default=1000, type=int)
    max_limit = max(1, int(config_module.TRACE_EXPORT_MAX_LIMIT))
    if limit > max_limit:
        limit = max_limit
    if limit <= 0:
        limit = 1

    filters = {
        "run_id": request.args.get("run_id"),
        "agent_id": request.args.get("agent_id"),
        "event_type": request.args.get("event_type"),
        "outcome": request.args.get("outcome"),
        "task_id": request.args.get("task_id"),
        "chunk_id": request.args.get("chunk_id"),
        "zone_id": request.args.get("zone_id"),
        "model_provider": request.args.get("model_provider"),
        "model_name": request.args.get("model_name"),
        "quality_status": request.args.get("quality_status"),
    }
    body = read_trace_events_ndjson(limit=limit, filters=filters)
    return Response(body, mimetype="application/x-ndjson")


@traces_bp.route("/model_routing_recommendation", methods=["GET"])
def model_routing_recommendation():
    try:
        _ = get_core_from_request()
    except PermissionError as e:
        return jsonify({"error": str(e)}), 403

    limit = request.args.get("limit", default=2000, type=int)
    if limit > 5000:
        limit = 5000
    if limit <= 0:
        limit = 1

    min_samples = request.args.get("min_samples", default=2, type=int)
    if min_samples <= 0:
        min_samples = 1

    routes_arg = str(request.args.get("routes") or "").strip()
    route_classes = [item.strip() for item in routes_arg.split(",") if item.strip()] if routes_arg else None

    events = read_trace_events(limit=limit, filters={})
    artifact = build_model_routing_recommendation(
        events=events,
        route_classes=route_classes,
        min_samples=min_samples,
    )
    artifact["event_window_count"] = len(events)
    artifact["limit"] = limit
    return jsonify(artifact)


@traces_bp.route("/model_routing_profile", methods=["GET"])
def model_routing_profile():
    try:
        _ = get_core_from_request()
    except PermissionError as e:
        return jsonify({"error": str(e)}), 403
    return jsonify(config_module.get_ai_route_model_profile())


@traces_bp.route("/model_routing_profile_apply", methods=["POST"])
def model_routing_profile_apply():
    try:
        _ = get_core_from_request()
    except PermissionError as e:
        return jsonify({"error": str(e)}), 403

    data = request.get_json(silent=True) or {}
    apply_change = bool(data.get("apply", False))
    default_model = data.get("default_model")
    route_models = dict(data.get("route_models") or {})
    artifact_id = str(data.get("artifact_id") or "").strip()

    artifact = data.get("artifact") if isinstance(data.get("artifact"), dict) else {}
    if artifact_id and not artifact:
        stored = get_model_routing_artifact(artifact_id)
        if not stored:
            return jsonify({"error": "artifact_not_found", "artifact_id": artifact_id}), 404
        artifact = dict(stored.get("artifact") or {})
    for item in list(artifact.get("recommendations") or []):
        if not isinstance(item, dict):
            continue
        route_class = str(item.get("route_class") or "").strip().lower()
        winner = item.get("winner") or {}
        model_name = str((winner.get("name") if isinstance(winner, dict) else "") or "").strip()
        if route_class and model_name and route_class not in route_models:
            route_models[route_class] = model_name

    before = config_module.get_ai_route_model_profile()
    preview = {
        "default_model": str(default_model).strip() if default_model is not None else before.get("default_model"),
        "routes": dict(before.get("routes") or {}),
    }
    for route, model in route_models.items():
        route_key = str(route or "").strip().lower()
        if route_key in preview["routes"]:
            preview["routes"][route_key] = str(model or "").strip()

    run_id = f"model-routing:profile:{request.headers.get('X-Session-ID') or 'none'}:{uuid.uuid4().hex[:8]}"
    actor = {
        "user_id": request.headers.get("X-User-ID"),
        "session_id": request.headers.get("X-Session-ID"),
    }
    emit_trace(
        build_trace_event(
            event_type="model_routing_profile_reviewed",
            run_id=run_id,
            outcome="success",
            payload={
                "apply_requested": apply_change,
                "artifact_id": artifact_id or None,
                "before": before,
                "preview": preview,
                "actor": actor,
            },
            agent_id=actor.get("user_id"),
        )
    )

    if not apply_change:
        return jsonify({"applied": False, "before": before, "preview": preview})

    after = config_module.apply_ai_route_model_profile(default_model=default_model, route_models=route_models)
    emit_trace(
        build_trace_event(
            event_type="model_routing_profile_applied",
            run_id=run_id,
            outcome="success",
            payload={
                "artifact_id": artifact_id or None,
                "before": before,
                "after": after,
                "actor": actor,
            },
            agent_id=actor.get("user_id"),
        )
    )
    return jsonify({"applied": True, "before": before, "after": after})


@traces_bp.route("/model_routing_artifacts", methods=["GET"])
def model_routing_artifacts_get():
    try:
        _ = get_core_from_request()
    except PermissionError as e:
        return jsonify({"error": str(e)}), 403

    artifact_id = str(request.args.get("artifact_id") or "").strip()
    if artifact_id:
        item = get_model_routing_artifact(artifact_id)
        if not item:
            return jsonify({"error": "artifact_not_found"}), 404
        return jsonify(item)

    limit = request.args.get("limit", default=20, type=int)
    if limit > 100:
        limit = 100
    if limit <= 0:
        limit = 1
    items = list_model_routing_artifacts(limit=limit)
    return jsonify({"artifacts": items, "count": len(items), "limit": limit})


@traces_bp.route("/model_routing_artifacts", methods=["POST"])
def model_routing_artifacts_post():
    try:
        _ = get_core_from_request()
    except PermissionError as e:
        return jsonify({"error": str(e)}), 403

    data = request.get_json(silent=True) or {}
    limit = int(data.get("limit") or 2000)
    if limit > 5000:
        limit = 5000
    if limit <= 0:
        limit = 1
    min_samples = int(data.get("min_samples") or 2)
    if min_samples <= 0:
        min_samples = 1
    route_classes = data.get("routes")
    if isinstance(route_classes, list):
        route_classes = [str(item).strip() for item in route_classes if str(item).strip()]
    else:
        route_classes = None

    events = read_trace_events(limit=limit, filters={})
    artifact = build_model_routing_recommendation(
        events=events,
        route_classes=route_classes,
        min_samples=min_samples,
    )
    artifact["event_window_count"] = len(events)
    artifact["limit"] = limit
    stored = create_model_routing_artifact(
        artifact=artifact,
        meta={
            "created_by_user": request.headers.get("X-User-ID"),
            "session_id": request.headers.get("X-Session-ID"),
        },
    )
    emit_trace(
        build_trace_event(
            event_type="model_routing_artifact_created",
            run_id=f"model-routing:artifact:{stored.get('artifact_id')}",
            outcome="success",
            payload={
                "artifact_id": stored.get("artifact_id"),
                "event_window_count": artifact.get("event_window_count"),
                "routes_evaluated": artifact.get("routes_evaluated"),
            },
            agent_id=request.headers.get("X-User-ID"),
        )
    )
    return jsonify(stored), 201



import time
from collections import defaultdict, deque
from threading import RLock

from hiveos.intelligence.contracts import validate_tool_request


def _normalize_caps(capabilities):
    if isinstance(capabilities, list):
        return capabilities
    if isinstance(capabilities, str):
        return [capabilities]
    return []


class ToolCapabilityRegistry:
    """
    Deny-by-default tool authorization and execution.
    """

    def __init__(self):
        self._lock = RLock()
        self._tools = {}
        self._call_windows = defaultdict(deque)  # (agent_id, tool_id) -> timestamps
        self._cost_windows = defaultdict(deque)  # (agent_id, tool_id) -> (ts, cost)
        self._decisions = {"allow": 0, "deny": 0}

    def register_tool(self, tool_id, capability, handler, request_schema=None, limits=None):
        with self._lock:
            self._tools[tool_id] = {
                "capability": capability,
                "handler": handler,
                "request_schema": request_schema or {},
                "limits": limits or {},
            }

    def authorize(self, agent, tool_request):
        request = validate_tool_request(tool_request)
        tool_id = request["tool_id"]

        with self._lock:
            tool = self._tools.get(tool_id)
            if not tool:
                self._decisions["deny"] += 1
                return {"allowed": False, "reason": "tool_not_registered"}

            required_cap = tool["capability"]
            agent_caps = _normalize_caps(agent.report_capabilities())
            if required_cap not in agent_caps:
                self._decisions["deny"] += 1
                return {"allowed": False, "reason": "capability_mismatch"}

            if not self._validate_schema(tool["request_schema"], request.get("args", {})):
                self._decisions["deny"] += 1
                return {"allowed": False, "reason": "schema_validation_failed"}

            if not self._within_limits(tool["limits"], request):
                self._decisions["deny"] += 1
                return {"allowed": False, "reason": "rate_or_budget_exceeded"}

            self._record_usage(request)
            self._decisions["allow"] += 1
            return {"allowed": True, "reason": "authorized"}

    def execute(self, tool_request, context=None):
        request = validate_tool_request(tool_request)
        with self._lock:
            tool = self._tools.get(request["tool_id"])
            if not tool:
                raise ValueError("Tool not registered")
            handler = tool["handler"]
        return handler(request.get("args", {}), context or {})

    def get_stats(self):
        with self._lock:
            return {
                "registered_tools": len(self._tools),
                "decisions": dict(self._decisions),
            }

    def _validate_schema(self, schema, args):
        required_args = schema.get("required_args", [])
        if not isinstance(args, dict):
            return False
        for key in required_args:
            if key not in args:
                return False
        return True

    def _within_limits(self, limits, request):
        agent_id = request["agent_id"]
        tool_id = request["tool_id"]
        now = time.time()

        # Calls per minute window
        max_calls = limits.get("max_calls_per_minute")
        call_key = (agent_id, tool_id)
        calls = self._call_windows[call_key]
        while calls and now - calls[0] > 60:
            calls.popleft()
        if max_calls is not None and len(calls) >= int(max_calls):
            return False

        # Cost per hour window
        max_cost = limits.get("max_cost_per_hour")
        cost_key = (agent_id, tool_id)
        costs = self._cost_windows[cost_key]
        while costs and now - costs[0][0] > 3600:
            costs.popleft()
        cost_now = sum(c for _, c in costs)
        req_cost = float(request.get("budget_cost", 0))
        if max_cost is not None and (cost_now + req_cost) > float(max_cost):
            return False

        return True

    def _record_usage(self, request):
        agent_id = request["agent_id"]
        tool_id = request["tool_id"]
        now = time.time()
        self._call_windows[(agent_id, tool_id)].append(now)
        self._cost_windows[(agent_id, tool_id)].append((now, float(request.get("budget_cost", 0))))




import unittest

from hiveos.services.intent_service import (
    SESSION_TERMINAL_TARGET_ID,
    _canonicalize_tool_and_action,
    _extract_command,
    _extract_requested_language,
    _fallback_plan,
    _worker_scope_for_target,
)


class IntentNormalizationTests(unittest.TestCase):
    def test_terminal_open_alias_normalizes(self):
        tool, action = _canonicalize_tool_and_action("terminal.open", "create", {})
        self.assertEqual("terminal.session", tool)
        self.assertEqual("open", action)

    def test_terminal_execute_alias_normalizes(self):
        tool, action = _canonicalize_tool_and_action("terminal.execute", "execute", {})
        self.assertEqual("terminal.exec", tool)
        self.assertEqual("run", action)

    def test_generic_terminal_with_run_action_normalizes(self):
        tool, action = _canonicalize_tool_and_action("terminal", "run", {"command": "echo hi"})
        self.assertEqual("terminal.exec", tool)
        self.assertEqual("run", action)

    def test_extract_command_preserves_echo_with_quoted_text(self):
        command = _extract_command('spin up a terminal and echo "hello friend. Welcome to HiveOS"')
        self.assertEqual('echo "hello friend. Welcome to HiveOS"', command)

    def test_fallback_plan_uses_session_terminal_when_requested(self):
        plan = _fallback_plan('use the session terminal and run echo "hello"')
        self.assertEqual("session-terminal", plan["target_id"])
        open_step = plan["steps"][0]
        self.assertEqual("terminal.session", open_step["tool_id"])
        self.assertEqual("session-terminal", open_step["args"]["target_id"])

    def test_fallback_plan_uses_scoped_terminal_by_default(self):
        plan = _fallback_plan('spin up a terminal and run echo "hello"')
        self.assertTrue(str(plan["target_id"]).startswith("intent-terminal-"))

    def test_worker_scope_for_session_target(self):
        self.assertEqual("session", _worker_scope_for_target(SESSION_TERMINAL_TARGET_ID))

    def test_worker_scope_for_scoped_target(self):
        self.assertEqual("scoped", _worker_scope_for_target("intent-terminal-abc123"))

    def test_extract_requested_language_french(self):
        language = _extract_requested_language("translate session terminal to french")
        self.assertEqual("french", language)

    def test_fallback_localize_uses_requested_language(self):
        plan = _fallback_plan("translate session terminal to french")
        transform_steps = [
            step for step in plan.get("steps", [])
            if step.get("tool_id") == "terminal.transform" and step.get("action") == "apply"
        ]
        self.assertTrue(transform_steps)
        config = transform_steps[0].get("args", {}).get("config", {})
        self.assertEqual("french", config.get("language"))


if __name__ == "__main__":
    unittest.main()



from collections import Counter

from hiveos.runtime.states import AgentStatus, ChunkStatus, TaskStatus, ZoneStallState


class ZoneWatchdog:
    def __init__(self, zone, stall_warn_ticks=200, stall_fail_ticks=0):
        self.zone = zone
        self.stall_warn_ticks = max(1, int(stall_warn_ticks))
        self.stall_fail_ticks = max(0, int(stall_fail_ticks))
        self.stall_ticks = 0
        self.blocked_since_tick = None
        self.blocked_reasons = []
        self.blocked_reason_counts = {}
        self.stall_state = ZoneStallState.HEALTHY

    def build_progress_fingerprint(self, task):
        chunks = self._task_chunks()
        status_counter = Counter(chunk.status for chunk in chunks)
        work = []
        for agent in self.zone.agent_instances:
            work.append(
                (
                    agent.agent_id,
                    agent.status,
                    agent.current_chunk,
                    bool(getattr(agent, "ready_to_execute", False)),
                    bool(getattr(agent, "waiting_on_input", False)),
                )
            )
        work.sort()
        return (
            task.status,
            tuple(sorted(status_counter.items())),
            tuple(work),
        )

    def propagate_stale_dependencies(self, pending_chunks):
        changed = False
        for chunk in pending_chunks:
            dep_id = chunk.depends_on
            if not dep_id:
                continue
            dep_chunk = self.zone.core.chunk_objects.get(dep_id)
            if dep_chunk and dep_chunk.status == ChunkStatus.STALE:
                chunk.mark_stale()
                changed = True
        return changed

    def refresh_blocked_reasons(self, task, pending_chunks):
        reasons = []
        for chunk in pending_chunks:
            reason = self._classify_pending_chunk_reason(task, chunk)
            reasons.append(reason)

        for agent in self.zone.agent_instances:
            if agent.status != AgentStatus.WORKING:
                continue
            if bool(getattr(agent, "waiting_on_input", False)):
                reasons.append("waiting_input")
            elif not bool(getattr(agent, "ready_to_execute", False)):
                reasons.append("execution_gate")

        if not reasons and pending_chunks:
            reasons.append("unknown_pending")

        self.blocked_reasons = reasons
        self.blocked_reason_counts = dict(Counter(reasons))

    def update_stall_state(self, task, pending_chunks, made_progress, current_tick):
        previous_state = self.stall_state

        if not pending_chunks:
            self.stall_ticks = 0
            self.blocked_since_tick = None
            self.stall_state = ZoneStallState.HEALTHY
            if task.status == TaskStatus.BLOCKED:
                task.status = TaskStatus.RUNNING
            return previous_state != self.stall_state

        if made_progress:
            self.stall_ticks = 0
            self.blocked_since_tick = None
            if task.status == TaskStatus.BLOCKED:
                task.status = TaskStatus.RUNNING
            self.stall_state = ZoneStallState.HEALTHY
            return previous_state != self.stall_state

        self.stall_ticks += 1
        if self.blocked_since_tick is None and current_tick is not None:
            self.blocked_since_tick = int(current_tick)

        if self.stall_ticks >= self.stall_warn_ticks:
            self.stall_state = ZoneStallState.BLOCKED
            if task.status in (TaskStatus.PENDING, TaskStatus.RUNNING):
                task.status = TaskStatus.BLOCKED
        else:
            self.stall_state = ZoneStallState.HEALTHY

        if self.stall_fail_ticks > 0 and self.stall_ticks >= self.stall_fail_ticks:
            for chunk in pending_chunks:
                chunk.mark_stale()
            task.status = TaskStatus.FAILED
            self.stall_state = ZoneStallState.FAILED
            self.zone.should_shutdown = True

        return previous_state != self.stall_state

    def build_tick_report(self, made_progress, state_changed):
        return {
            "zone_id": self.zone.zone_id,
            "task_id": self.zone.task_id,
            "made_progress": bool(made_progress),
            "stall_state": self.stall_state,
            "stall_ticks": int(self.stall_ticks),
            "state_changed": bool(state_changed),
            "blocked_reasons": list(self.blocked_reasons),
            "blocked_reason_counts": dict(self.blocked_reason_counts),
        }

    def get_summary(self):
        return {
            "state": self.stall_state,
            "stall_ticks": int(self.stall_ticks),
            "warn_ticks": int(self.stall_warn_ticks),
            "fail_ticks": int(self.stall_fail_ticks),
            "blocked_since_tick": self.blocked_since_tick,
            "blocked_reasons": list(self.blocked_reasons),
            "blocked_reason_counts": dict(self.blocked_reason_counts),
        }

    def _pending_chunks(self):
        return [
            c for c in self.zone.core.chunk_objects.values()
            if c.task_id == self.zone.task_id and c.status == ChunkStatus.PENDING
        ]

    def _task_chunks(self):
        return [c for c in self.zone.core.chunk_objects.values() if c.task_id == self.zone.task_id]

    def _classify_pending_chunk_reason(self, task, chunk):
        dep_id = chunk.depends_on
        if dep_id:
            dep_chunk = self.zone.core.chunk_objects.get(dep_id)
            if dep_chunk is None:
                return "dependency_missing"
            if dep_chunk.status == ChunkStatus.STALE:
                return "dependency_stale"
            if dep_chunk.status != ChunkStatus.COMPLETED:
                return "dependency_pending"

        cond = chunk.conditional
        if task.has_conditionals and isinstance(cond, dict) and cond.get("type"):
            key = str(cond.get("key") or "").strip()
            if key and self.zone.local_bus.get(key) is None:
                return "conditional_unresolved"
            return "conditional_blocked"

        cap = str(chunk.required_capability or "generic")
        if not self._capability_exists_in_session(cap):
            return f"missing_capability:{cap}"
        if not self._idle_capability_available_in_zone(cap):
            return f"waiting_capability_agent:{cap}"
        return "assignable_pending"

    def _capability_exists_in_session(self, capability):
        for runtime_agent in self.zone.core._runtime_agents.values():
            if runtime_agent.session_id != self.zone.session_id:
                continue
            if capability in list(getattr(runtime_agent.obj, "capabilities", []) or []):
                return True
        return False

    def _idle_capability_available_in_zone(self, capability):
        return any(
            agent.status == AgentStatus.IDLE and capability in list(getattr(agent, "capabilities", []) or [])
            for agent in self.zone.agent_instances
        )



# messaging/message_bus.py

from threading import Lock

class MessageBus:
    def __init__(self):
        self._lock = Lock()
        self._store = {}
        self._subscribers = []

    def publish(self, key, value):
        callbacks = []
        with self._lock:
            self._store[key] = value
            for pattern, callback in self._subscribers:
                if self._match(pattern, key):
                    callbacks.append(callback)
        for callback in callbacks:
            try:
                callback(key, value)
            except Exception:
                # Isolate bus publisher from subscriber failures.
                continue

    def subscribe(self, pattern, callback):
        with self._lock:
            self._subscribers.append((pattern, callback))

    def _match(self, pattern, key):
        # Simple "+" wildcard matching
        pattern_parts = pattern.split("/")
        key_parts = key.split("/")

        if len(pattern_parts) != len(key_parts):
            return False

        for p, k in zip(pattern_parts, key_parts):
            if p == "+":
                continue
            if p != k:
                return False
        return True

    def get(self, key, default=None):
        with self._lock:
            return self._store.get(key, default)

    def snapshot(self):
        with self._lock:
            return self._store.copy()

    def clear(self):
        with self._lock:
            self._store.clear()

    def keys(self):
        with self._lock:
            return list(self._store.keys())

    def items(self):
        with self._lock:
            return list(self._store.items())



import json
import os
import queue
import random
import threading
import time
import uuid
from collections import deque
from datetime import datetime, timezone

from hiveos.config import (
    TRACE_DB_ASYNC,
    TRACE_DB_QUEUE_MAX,
    TRACE_LOG_PATH,
    TRACE_MAX_BYTES,
    TRACE_MAX_ROTATIONS,
    TRACE_READ_MAX_LINES,
    TRACE_SAMPLING_ALWAYS_EVENT_TYPES,
    TRACE_SAMPLING_BASE_RATE,
    TRACE_SAMPLING_DEGRADED_QUEUE_PCT,
    TRACE_SAMPLING_DEGRADED_RATE,
    TRACE_SAMPLING_ENABLED,
    TRACE_SAMPLING_SEVERE_QUEUE_PCT,
    TRACE_SAMPLING_SEVERE_RATE,
    TRACE_SAMPLING_AUTOTUNE_DROP_RATIO_PCT,
    TRACE_SAMPLING_AUTOTUNE_ENABLED,
    TRACE_SAMPLING_AUTOTUNE_HIGH_QUEUE_PCT,
    TRACE_SAMPLING_AUTOTUNE_INTERVAL_SEC,
    TRACE_SAMPLING_AUTOTUNE_MIN_RATE,
    TRACE_SAMPLING_AUTOTUNE_STEP,
    TRACE_SAMPLING_AUTOTUNE_TARGET_QUEUE_PCT,
)
from hiveos.intelligence.contracts import validate_trace_event
from hiveos.metrics import log_event

_TRACE_LOCK = threading.Lock()
_TRACE_DB_QUEUE = queue.Queue(maxsize=max(1, TRACE_DB_QUEUE_MAX))
_TRACE_DB_WORKER = None
_TRACE_DB_WORKER_LOCK = threading.Lock()
_TRACE_STATS_LOCK = threading.Lock()
_TRACE_STATS = {
    "total": 0,
    "sampled_out": 0,
    "persisted": 0,
    "db_enqueued": 0,
    "db_dropped": 0,
    "db_written": 0,
    "db_failed": 0,
}
_TRACE_AUTOTUNE_LOCK = threading.Lock()
_TRACE_AUTOTUNE_STATE = {
    "degraded_rate": float(TRACE_SAMPLING_DEGRADED_RATE),
    "severe_rate": float(TRACE_SAMPLING_SEVERE_RATE),
    "last_eval_monotonic": 0.0,
    "last_db_enqueued": 0,
    "last_db_dropped": 0,
    "last_drop_ratio": 0.0,
    "last_queue_pct": 0,
    "last_action": "init",
}


def _ensure_trace_log_parent():
    try:
        path = os.path.abspath(str(TRACE_LOG_PATH or "trace_events.log"))
        parent = os.path.dirname(path)
        if parent:
            os.makedirs(parent, exist_ok=True)
    except OSError:
        return


def now_iso():
    return datetime.now(timezone.utc).isoformat()


def build_trace_event(event_type, run_id, outcome="success", payload=None, **kwargs):
    model_meta = _normalize_model_meta(kwargs.get("model"))
    event = {
        "trace_id": kwargs.get("trace_id") or str(uuid.uuid4()),
        "span_id": kwargs.get("span_id") or str(uuid.uuid4())[:12],
        "parent_span_id": kwargs.get("parent_span_id"),
        "run_id": run_id,
        "timestamp": kwargs.get("timestamp") or now_iso(),
        "event_type": event_type,
        "agent_id": kwargs.get("agent_id"),
        "task_id": kwargs.get("task_id"),
        "chunk_id": kwargs.get("chunk_id"),
        "zone_id": kwargs.get("zone_id"),
        "payload": payload or {},
        "timing": kwargs.get("timing") or {},
        "model": model_meta,
        "outcome": outcome,
    }
    return validate_trace_event(event)


def _normalize_model_meta(model_meta):
    if not isinstance(model_meta, dict):
        return {}
    normalized = dict(model_meta)
    name = normalized.get("name")
    legacy = normalized.get("model")
    if not isinstance(name, str) or not name.strip():
        if isinstance(legacy, str) and legacy.strip():
            normalized["name"] = legacy.strip()
    provider = normalized.get("provider")
    if isinstance(provider, str):
        normalized["provider"] = provider.strip()
    if isinstance(normalized.get("name"), str):
        normalized["name"] = normalized["name"].strip()
    return normalized


def emit_trace(event, persist_db=True, persist_file=True):
    validated = validate_trace_event(event)
    _increment_trace_stat("total")
    _maybe_auto_tune()

    sample = _sample_decision(validated)
    if not sample["keep"]:
        _increment_trace_stat("sampled_out")
        return validated

    if persist_file:
        line = json.dumps(validated, separators=(",", ":")) + "\n"
        _ensure_trace_log_parent()
        with _TRACE_LOCK:
            _rotate_trace_files_if_needed(incoming_bytes=len(line.encode("utf-8")))
            with open(TRACE_LOG_PATH, "a", encoding="utf-8") as handle:
                handle.write(line)

    if persist_db:
        if TRACE_DB_ASYNC:
            _ensure_db_worker_started()
            job = {
                "agent_id": validated.get("agent_id"),
                "chunk_id": validated.get("chunk_id"),
                "event_type": validated.get("event_type"),
                "zone": validated.get("zone_id"),
                "task_id": validated.get("task_id"),
                "result": validated.get("outcome"),
                "reason": json.dumps(validated.get("payload", {}), separators=(",", ":")),
                "tick": None,
            }
            try:
                _TRACE_DB_QUEUE.put_nowait(job)
                _increment_trace_stat("db_enqueued")
            except queue.Full:
                _increment_trace_stat("db_dropped")
        else:
            _write_trace_to_db(
                agent_id=validated.get("agent_id"),
                chunk_id=validated.get("chunk_id"),
                event_type=validated.get("event_type"),
                zone=validated.get("zone_id"),
                task_id=validated.get("task_id"),
                result=validated.get("outcome"),
                reason=json.dumps(validated.get("payload", {}), separators=(",", ":")),
                tick=None,
            )

    _increment_trace_stat("persisted")
    return validated


def read_trace_events(limit=200, filters=None):
    filters = filters or {}
    if limit <= 0:
        return []
    matched = deque(maxlen=limit)
    scanned = 0

    for path in _iter_trace_files_oldest_first():
        try:
            with open(path, "r", encoding="utf-8") as handle:
                for line in handle:
                    if scanned >= TRACE_READ_MAX_LINES:
                        return list(matched)
                    scanned += 1

                    line = line.strip()
                    if not line:
                        continue
                    try:
                        event = json.loads(line)
                    except json.JSONDecodeError:
                        continue
                    if _matches_filters(event, filters):
                        matched.append(event)
        except OSError:
            continue

    return list(matched)


def read_trace_events_ndjson(limit=200, filters=None):
    events = read_trace_events(limit=limit, filters=filters)
    if not events:
        return ""
    return "\n".join(json.dumps(event, separators=(",", ":")) for event in events) + "\n"


def _matches_filters(event, filters):
    for key in ("run_id", "agent_id", "event_type", "outcome", "task_id", "chunk_id", "zone_id"):
        wanted = filters.get(key)
        if wanted is None:
            continue
        if str(event.get(key)) != str(wanted):
            return False
    model = event.get("model") or {}
    payload = event.get("payload") or {}
    wanted_provider = filters.get("model_provider")
    if wanted_provider is not None and str(model.get("provider")) != str(wanted_provider):
        return False
    wanted_name = filters.get("model_name")
    if wanted_name is not None and str(model.get("name")) != str(wanted_name):
        return False
    wanted_quality = filters.get("quality_status")
    if wanted_quality is not None and str(payload.get("quality_status")) != str(wanted_quality):
        return False
    return True


def _rotate_trace_files_if_needed(incoming_bytes):
    if TRACE_MAX_BYTES <= 0:
        return
    try:
        current_size = os.path.getsize(TRACE_LOG_PATH)
    except OSError:
        current_size = 0

    if current_size + incoming_bytes <= TRACE_MAX_BYTES:
        return

    max_rotations = max(0, int(TRACE_MAX_ROTATIONS))
    if max_rotations == 0:
        try:
            os.remove(TRACE_LOG_PATH)
        except OSError:
            pass
        return

    oldest = f"{TRACE_LOG_PATH}.{max_rotations}"
    if os.path.exists(oldest):
        try:
            os.remove(oldest)
        except OSError:
            pass

    for idx in range(max_rotations - 1, 0, -1):
        src = f"{TRACE_LOG_PATH}.{idx}"
        dst = f"{TRACE_LOG_PATH}.{idx + 1}"
        if os.path.exists(src):
            try:
                os.replace(src, dst)
            except OSError:
                continue

    if os.path.exists(TRACE_LOG_PATH):
        try:
            os.replace(TRACE_LOG_PATH, f"{TRACE_LOG_PATH}.1")
        except OSError:
            pass


def _iter_trace_files_oldest_first():
    max_rotations = max(0, int(TRACE_MAX_ROTATIONS))
    for idx in range(max_rotations, 0, -1):
        candidate = f"{TRACE_LOG_PATH}.{idx}"
        if os.path.exists(candidate):
            yield candidate
    if os.path.exists(TRACE_LOG_PATH):
        yield TRACE_LOG_PATH


def get_trace_sink_stats():
    with _TRACE_STATS_LOCK:
        stats = dict(_TRACE_STATS)
    queue_depth = _TRACE_DB_QUEUE.qsize() if TRACE_DB_ASYNC else 0
    try:
        current_log_bytes = os.path.getsize(TRACE_LOG_PATH)
    except OSError:
        current_log_bytes = 0
    queue_pct = _queue_pressure_pct()
    autotune = _get_autotune_state()
    active_rates = _get_active_rates()
    sample = _sample_decision(None)
    return {
        "file_path": TRACE_LOG_PATH,
        "file_bytes": current_log_bytes,
        "max_bytes": TRACE_MAX_BYTES,
        "max_rotations": TRACE_MAX_ROTATIONS,
        "read_max_lines": TRACE_READ_MAX_LINES,
        "db_async": TRACE_DB_ASYNC,
        "db_queue_max": TRACE_DB_QUEUE_MAX,
        "db_queue_depth": queue_depth,
        "db_queue_pressure_pct": queue_pct,
        "sampling": {
            "enabled": TRACE_SAMPLING_ENABLED,
            "mode": sample["mode"],
            "active_rate": sample["rate"],
            "base_rate": _clamp_rate(TRACE_SAMPLING_BASE_RATE),
            "degraded_rate": active_rates["degraded_rate"],
            "severe_rate": active_rates["severe_rate"],
            "degraded_queue_pct": TRACE_SAMPLING_DEGRADED_QUEUE_PCT,
            "severe_queue_pct": TRACE_SAMPLING_SEVERE_QUEUE_PCT,
            "always_event_types": list(TRACE_SAMPLING_ALWAYS_EVENT_TYPES),
            "always_outcomes": ["denied", "failed"],
            "autotune": autotune,
        },
        "stats": stats,
    }


def _ensure_db_worker_started():
    global _TRACE_DB_WORKER
    if _TRACE_DB_WORKER is not None and _TRACE_DB_WORKER.is_alive():
        return
    with _TRACE_DB_WORKER_LOCK:
        if _TRACE_DB_WORKER is not None and _TRACE_DB_WORKER.is_alive():
            return
        worker = threading.Thread(target=_db_worker_loop, daemon=True, name="hive-trace-db-writer")
        worker.start()
        _TRACE_DB_WORKER = worker


def _db_worker_loop():
    while True:
        job = _TRACE_DB_QUEUE.get()
        try:
            _write_trace_to_db(**job)
        finally:
            _TRACE_DB_QUEUE.task_done()


def _write_trace_to_db(agent_id, chunk_id, event_type, zone, task_id, result, reason, tick):
    try:
        log_event(
            agent_id=agent_id,
            chunk_id=chunk_id,
            event_type=event_type,
            zone=zone,
            task_id=task_id,
            result=result,
            reason=reason,
            tick=tick,
        )
        _increment_trace_stat("db_written")
    except Exception:
        _increment_trace_stat("db_failed")


def _increment_trace_stat(key):
    with _TRACE_STATS_LOCK:
        _TRACE_STATS[key] = int(_TRACE_STATS.get(key, 0)) + 1


def _sample_decision(event):
    if not TRACE_SAMPLING_ENABLED:
        return {"keep": True, "mode": "disabled", "rate": 1.0}

    if event is not None:
        outcome = str(event.get("outcome") or "").lower()
        event_type = str(event.get("event_type") or "")
        if outcome in {"denied", "failed"}:
            return {"keep": True, "mode": "critical", "rate": 1.0}
        if event_type in set(TRACE_SAMPLING_ALWAYS_EVENT_TYPES):
            return {"keep": True, "mode": "critical", "rate": 1.0}

    queue_pct = _queue_pressure_pct()
    active_rates = _get_active_rates()
    mode = "normal"
    rate = active_rates["base_rate"]
    if queue_pct >= int(TRACE_SAMPLING_SEVERE_QUEUE_PCT):
        mode = "severe"
        rate = active_rates["severe_rate"]
    elif queue_pct >= int(TRACE_SAMPLING_DEGRADED_QUEUE_PCT):
        mode = "degraded"
        rate = active_rates["degraded_rate"]

    if rate >= 1.0:
        return {"keep": True, "mode": mode, "rate": rate}
    return {"keep": random.random() <= rate, "mode": mode, "rate": rate}


def _queue_pressure_pct():
    if not TRACE_DB_ASYNC or TRACE_DB_QUEUE_MAX <= 0:
        return 0
    try:
        depth = _TRACE_DB_QUEUE.qsize()
    except Exception:
        depth = 0
    return int((depth / max(1, TRACE_DB_QUEUE_MAX)) * 100)


def _clamp_rate(value):
    return max(0.0, min(1.0, float(value)))


def _get_active_rates():
    base_rate = _clamp_rate(TRACE_SAMPLING_BASE_RATE)
    if not TRACE_SAMPLING_AUTOTUNE_ENABLED:
        return {
            "base_rate": base_rate,
            "degraded_rate": _clamp_rate(TRACE_SAMPLING_DEGRADED_RATE),
            "severe_rate": _clamp_rate(TRACE_SAMPLING_SEVERE_RATE),
        }
    with _TRACE_AUTOTUNE_LOCK:
        degraded = _clamp_rate(_TRACE_AUTOTUNE_STATE["degraded_rate"])
        severe = _clamp_rate(_TRACE_AUTOTUNE_STATE["severe_rate"])
    if severe > degraded:
        severe = degraded
    return {
        "base_rate": base_rate,
        "degraded_rate": degraded,
        "severe_rate": severe,
    }


def _get_autotune_state():
    with _TRACE_AUTOTUNE_LOCK:
        state = dict(_TRACE_AUTOTUNE_STATE)
    return {
        "enabled": TRACE_SAMPLING_AUTOTUNE_ENABLED,
        "interval_sec": max(1, int(TRACE_SAMPLING_AUTOTUNE_INTERVAL_SEC)),
        "step": _clamp_rate(TRACE_SAMPLING_AUTOTUNE_STEP),
        "min_rate": _clamp_rate(TRACE_SAMPLING_AUTOTUNE_MIN_RATE),
        "target_queue_pct": int(TRACE_SAMPLING_AUTOTUNE_TARGET_QUEUE_PCT),
        "high_queue_pct": int(TRACE_SAMPLING_AUTOTUNE_HIGH_QUEUE_PCT),
        "drop_ratio_pct": float(TRACE_SAMPLING_AUTOTUNE_DROP_RATIO_PCT),
        "last_action": state.get("last_action"),
        "last_drop_ratio": round(float(state.get("last_drop_ratio", 0.0)), 4),
        "last_queue_pct": int(state.get("last_queue_pct", 0)),
        "last_eval_monotonic": round(float(state.get("last_eval_monotonic", 0.0)), 3),
    }


def _maybe_auto_tune():
    if not TRACE_SAMPLING_AUTOTUNE_ENABLED:
        return

    now_mono = time.monotonic()
    interval = max(1, int(TRACE_SAMPLING_AUTOTUNE_INTERVAL_SEC))
    queue_pct = _queue_pressure_pct()

    with _TRACE_AUTOTUNE_LOCK:
        last_eval = float(_TRACE_AUTOTUNE_STATE.get("last_eval_monotonic", 0.0))
        if now_mono - last_eval < interval:
            return
        _TRACE_AUTOTUNE_STATE["last_eval_monotonic"] = now_mono
        _TRACE_AUTOTUNE_STATE["last_queue_pct"] = int(queue_pct)

    with _TRACE_STATS_LOCK:
        enqueued = int(_TRACE_STATS.get("db_enqueued", 0))
        dropped = int(_TRACE_STATS.get("db_dropped", 0))

    with _TRACE_AUTOTUNE_LOCK:
        prev_enqueued = int(_TRACE_AUTOTUNE_STATE.get("last_db_enqueued", 0))
        prev_dropped = int(_TRACE_AUTOTUNE_STATE.get("last_db_dropped", 0))
        delta_enqueued = max(0, enqueued - prev_enqueued)
        delta_dropped = max(0, dropped - prev_dropped)
        denom = max(1, delta_enqueued + delta_dropped)
        drop_ratio = float(delta_dropped) / float(denom)
        _TRACE_AUTOTUNE_STATE["last_db_enqueued"] = enqueued
        _TRACE_AUTOTUNE_STATE["last_db_dropped"] = dropped
        _TRACE_AUTOTUNE_STATE["last_drop_ratio"] = drop_ratio

        base_degraded = _clamp_rate(TRACE_SAMPLING_DEGRADED_RATE)
        base_severe = _clamp_rate(TRACE_SAMPLING_SEVERE_RATE)
        step = max(0.001, _clamp_rate(TRACE_SAMPLING_AUTOTUNE_STEP))
        min_rate = _clamp_rate(TRACE_SAMPLING_AUTOTUNE_MIN_RATE)
        high_queue = int(TRACE_SAMPLING_AUTOTUNE_HIGH_QUEUE_PCT)
        target_queue = int(TRACE_SAMPLING_AUTOTUNE_TARGET_QUEUE_PCT)
        drop_ratio_threshold = float(TRACE_SAMPLING_AUTOTUNE_DROP_RATIO_PCT) / 100.0

        degraded = _clamp_rate(_TRACE_AUTOTUNE_STATE.get("degraded_rate", base_degraded))
        severe = _clamp_rate(_TRACE_AUTOTUNE_STATE.get("severe_rate", base_severe))

        if queue_pct >= high_queue or drop_ratio >= drop_ratio_threshold:
            degraded = max(min_rate, degraded - step)
            severe = max(min_rate, severe - step)
            _TRACE_AUTOTUNE_STATE["last_action"] = "tighten"
        elif queue_pct <= target_queue and drop_ratio <= 0.0:
            recover_step = step * 0.5
            degraded = min(base_degraded, degraded + recover_step)
            severe = min(base_severe, severe + recover_step)
            _TRACE_AUTOTUNE_STATE["last_action"] = "relax"
        else:
            _TRACE_AUTOTUNE_STATE["last_action"] = "hold"

        if severe > degraded:
            severe = degraded
        _TRACE_AUTOTUNE_STATE["degraded_rate"] = degraded
        _TRACE_AUTOTUNE_STATE["severe_rate"] = severe



import time
from hiveos.metrics import log_event
from hiveos.core import Core

core = Core()

def ingest_intent(intent):
    core.ingest_intent(intent)

def tick(verbose=True):
    # start = time.perf_counter()
    # print("⏱ Tick started...")
    # print()

    core.tick(verbose=verbose)

    # log_event("system", None, "tick_completed", "global")

    # end = time.perf_counter()
    # elapsed_ms = (end - start) * 1000
    # print(f"✅ Tick complete in {elapsed_ms:.2f} ms.")



import tempfile
import unittest
from pathlib import Path
import uuid

from hiveos import db as db_module
from hiveos.config import USER_CAPABILITIES_HEADER
from hiveos.db import init_db
from hiveos.intelligence import tracing as tracing_module
from hiveos.session_manager import session_manager
from web_interface import app


class ActionTraceOutcomeTests(unittest.TestCase):
    def setUp(self):
        self._original_trace_db_async = tracing_module.TRACE_DB_ASYNC
        self._original_trace_sampling_enabled = tracing_module.TRACE_SAMPLING_ENABLED
        self._original_trace_log_path = tracing_module.TRACE_LOG_PATH

        self._db_path = Path(tempfile.gettempdir()) / f"hive-test-{uuid.uuid4().hex}.db"
        db_module.DB_PATH = str(self._db_path)
        init_db()

        self._trace_log_path = str(Path(tempfile.gettempdir()) / f"hive-trace-{uuid.uuid4().hex}.log")
        tracing_module.TRACE_LOG_PATH = self._trace_log_path
        tracing_module.TRACE_DB_ASYNC = False
        tracing_module.TRACE_SAMPLING_ENABLED = False

        self._cleanup_sessions()
        self.client = app.test_client()

    def tearDown(self):
        self._cleanup_sessions()
        tracing_module.TRACE_DB_ASYNC = self._original_trace_db_async
        tracing_module.TRACE_SAMPLING_ENABLED = self._original_trace_sampling_enabled
        tracing_module.TRACE_LOG_PATH = self._original_trace_log_path
        try:
            db_path = Path(self._db_path)
            if db_path.exists():
                db_path.unlink()
        except Exception:
            pass
        self._cleanup_trace_files()

    def _cleanup_sessions(self):
        session_ids = list(session_manager.sessions.keys())
        for session_id in session_ids:
            try:
                session_manager.delete_session(session_id)
            except Exception:
                pass

    def _cleanup_trace_files(self):
        base = Path(self._trace_log_path)
        for path in [base] + list(base.parent.glob(f"{base.name}.*")):
            try:
                if path.exists():
                    path.unlink()
            except Exception:
                pass

    def _headers(self, user_id, session_id=None):
        headers = {
            "X-User-ID": user_id,
            USER_CAPABILITIES_HEADER: "admin",
        }
        if session_id:
            headers["X-Session-ID"] = session_id
        return headers

    def _create_session(self, user_id):
        response = self.client.post(
            "/create_session",
            json={"label": "trace-test", "user_id": user_id},
            headers=self._headers(user_id),
        )
        self.assertEqual(200, response.status_code)
        return response.get_json()["session_id"]

    def _find_latest_event(self, event_type, action):
        events = tracing_module.read_trace_events(limit=500, filters={"event_type": event_type})
        for event in reversed(events):
            payload = event.get("payload") or {}
            if payload.get("action") == action:
                return event
        return None

    def test_create_session_emits_action_applied(self):
        user_id = "trace-user-create"
        session_id = self._create_session(user_id)

        event = self._find_latest_event("action_applied", "sessions.create")
        self.assertIsNotNone(event)
        payload = event["payload"]
        self.assertEqual("session", payload.get("resource_type"))
        self.assertEqual(session_id, payload.get("resource_id"))

    def test_runtime_inject_yaml_emits_action_applied(self):
        user_id = "trace-user-inject"
        session_id = self._create_session(user_id)
        yaml_payload = (
            "task_name: trace_task\n"
            "execution_mode: concurrent\n"
            "chunks:\n"
            "  - chunk_id: c1\n"
            "    required_capability: compute.generic\n"
            "    payload: {}\n"
        )

        response = self.client.post(
            "/inject_yaml",
            data=yaml_payload,
            headers=self._headers(user_id, session_id),
        )
        self.assertEqual(200, response.status_code)

        event = self._find_latest_event("action_applied", "runtime.inject_yaml")
        self.assertIsNotNone(event)
        payload = event["payload"]
        self.assertEqual("session", payload.get("resource_type"))
        self.assertIsNotNone(payload.get("task_id"))

    def test_delete_task_missing_emits_action_failed(self):
        user_id = "trace-user-delete"
        session_id = self._create_session(user_id)
        missing_task_id = 999999

        response = self.client.post(
            "/delete_task",
            json={"task_id": missing_task_id},
            headers=self._headers(user_id, session_id),
        )
        self.assertEqual(404, response.status_code)

        event = self._find_latest_event("action_failed", "tasks.delete")
        self.assertIsNotNone(event)
        payload = event["payload"]
        self.assertEqual("task", payload.get("resource_type"))
        self.assertEqual(missing_task_id, payload.get("resource_id"))
        self.assertEqual("task_not_found", payload.get("reason"))

    def test_inject_agents_for_pending_emits_action_applied(self):
        user_id = "trace-user-agent-pending"
        session_id = self._create_session(user_id)

        response = self.client.post(
            "/inject_agents_for_pending",
            json={"max_total": 2, "per_capability": 1},
            headers=self._headers(user_id, session_id),
        )
        self.assertEqual(200, response.status_code)

        event = self._find_latest_event("action_applied", "agents.inject_pending")
        self.assertIsNotNone(event)
        payload = event["payload"]
        self.assertEqual("session", payload.get("resource_type"))
        self.assertIn("injected_total", payload)

    def test_freeze_zone_missing_zone_emits_action_failed(self):
        user_id = "trace-user-freeze-zone"
        session_id = self._create_session(user_id)
        missing_zone_id = "zone-missing"

        response = self.client.post(
            "/freeze_zone",
            json={"zone_id": missing_zone_id},
            headers=self._headers(user_id, session_id),
        )
        self.assertEqual(404, response.status_code)

        event = self._find_latest_event("action_failed", "control.freeze_zone")
        self.assertIsNotNone(event)
        payload = event["payload"]
        self.assertEqual("zone", payload.get("resource_type"))
        self.assertEqual(missing_zone_id, payload.get("resource_id"))
        self.assertEqual("zone_not_found", payload.get("reason"))


if __name__ == "__main__":
    unittest.main()



import random
from hiveos.runtime.agent_runtime import agent_objects
from hiveos.runtime.chunk_runtime import get_task_id, reset

def induce_failure(agent_name=None):
    """
    If agent_name is provided, attempt to fail that specific agent.
    If None, fail a random working agent.
    """
    target_agent = None

    if agent_name:
        target_agent = agent_objects.get(agent_name)
        if not target_agent:
            print(f"[Demo] Agent '{agent_name}' not found in runtime.")
            return
    else:
        working_agents = [a for a in agent_objects.values() if a.status == 'working']
        if not working_agents:
            print("[Demo] No agents are currently working.")
            return
        target_agent = random.choice(working_agents)

    if target_agent.status == 'working' and target_agent.current_chunk:
        task_id = get_task_id(target_agent.current_chunk)
        reset(target_agent.current_chunk)

        target_agent.status = 'recharging'
        target_agent.cooldown_ticks = 5
        target_agent.current_chunk = None
        target_agent.chunk_progress = 0
        target_agent.required_ticks = 0
        print(f"[Demo] Induced failure on {target_agent.name}. Chunk reassigned. Agent now recharging.")
    else:
        print(f"[Demo] Agent {target_agent.name} is not working — no failure induced.")



import threading
import time
import uuid
from collections import deque
from datetime import datetime, timezone


def _now_iso() -> str:
    return datetime.now(timezone.utc).isoformat()


class AIRunState:
    QUEUED = "queued"
    RUNNING = "running"
    COMPLETED = "completed"
    FAILED = "failed"
    TIMEOUT = "timeout"
    DROPPED = "dropped"


class AIExecutionPlane:
    """
    Minimal async run-plane skeleton.
    This is intentionally simple and acts as the contract anchor for future model workers.
    """

    def __init__(self, max_runs=1000, max_queue=300, queue_ttl_seconds=30, max_runtime_seconds=20):
        self._lock = threading.Lock()
        self._queue = deque()
        self._runs = {}
        self._max_runs = max_runs
        self._max_queue = max(1, int(max_queue))
        self._queue_ttl_seconds = max(1, int(queue_ttl_seconds))
        self._max_runtime_seconds = max(1, int(max_runtime_seconds))
        self._worker = None
        self._running = False
        self._stats = {
            "submitted": 0,
            "completed": 0,
            "failed": 0,
            "timed_out": 0,
            "dropped_queue_full": 0,
            "dropped_queue_ttl": 0,
            "dropped_trimmed": 0,
        }

    def start(self):
        with self._lock:
            if self._running:
                return
            self._running = True
        self._worker = threading.Thread(target=self._worker_loop, daemon=True)
        self._worker.start()

    def stop(self):
        with self._lock:
            self._running = False

    def submit_run(self, agent_id, task_id, chunk_id, payload):
        run_id = f"{chunk_id}:{uuid.uuid4().hex[:8]}"
        run = {
            "run_id": run_id,
            "agent_id": agent_id,
            "task_id": task_id,
            "chunk_id": chunk_id,
            "state": AIRunState.QUEUED,
            "submitted_at": _now_iso(),
            "started_at": None,
            "finished_at": None,
            "payload": payload,
            "output": None,
            "error": None,
        }
        with self._lock:
            if len(self._queue) >= self._max_queue:
                run["state"] = AIRunState.DROPPED
                run["finished_at"] = _now_iso()
                run["error"] = "queue_full"
                self._runs[run_id] = run
                self._stats["dropped_queue_full"] += 1
                self._trim_old_runs_locked()
                return run_id
            self._runs[run_id] = run
            self._queue.append(run_id)
            self._stats["submitted"] += 1
            self._trim_old_runs_locked()
        return run_id

    def get_run(self, run_id):
        with self._lock:
            run = self._runs.get(run_id)
            return dict(run) if run else None

    def get_stats(self):
        with self._lock:
            queued = sum(1 for r in self._runs.values() if r["state"] == AIRunState.QUEUED)
            running = sum(1 for r in self._runs.values() if r["state"] == AIRunState.RUNNING)
            timed_out = sum(1 for r in self._runs.values() if r["state"] == AIRunState.TIMEOUT)
            dropped = sum(1 for r in self._runs.values() if r["state"] == AIRunState.DROPPED)
            return {
                "total_runs": len(self._runs),
                "queued_runs": queued,
                "running_runs": running,
                "timed_out_runs": timed_out,
                "dropped_runs": dropped,
                "queue_depth": len(self._queue),
                "max_queue": self._max_queue,
                "queue_ttl_seconds": self._queue_ttl_seconds,
                "max_runtime_seconds": self._max_runtime_seconds,
                "stats": dict(self._stats),
            }

    def _worker_loop(self):
        while True:
            with self._lock:
                running = self._running
                run_id = self._queue.popleft() if self._queue else None
            if not running:
                return
            if not run_id:
                time.sleep(0.01)
                continue
            self._execute_run(run_id)

    def _execute_run(self, run_id):
        with self._lock:
            run = self._runs.get(run_id)
            if not run:
                return
            age_seconds = _age_seconds(run.get("submitted_at"))
            if age_seconds > self._queue_ttl_seconds:
                run["state"] = AIRunState.DROPPED
                run["error"] = "queue_ttl_expired"
                run["finished_at"] = _now_iso()
                self._stats["dropped_queue_ttl"] += 1
                return
            run["state"] = AIRunState.RUNNING
            run["started_at"] = _now_iso()
            payload = run["payload"]
        exec_started = time.monotonic()

        try:
            # Placeholder async model execution path.
            output = payload
            with self._lock:
                run = self._runs.get(run_id)
                if not run:
                    return
                elapsed = time.monotonic() - exec_started
                if elapsed > self._max_runtime_seconds:
                    run["state"] = AIRunState.TIMEOUT
                    run["error"] = "run_timeout_exceeded"
                    run["finished_at"] = _now_iso()
                    self._stats["timed_out"] += 1
                    return
                run["output"] = output
                run["state"] = AIRunState.COMPLETED
                run["finished_at"] = _now_iso()
                self._stats["completed"] += 1
        except Exception as exc:  # pragma: no cover - defensive
            with self._lock:
                run = self._runs.get(run_id)
                if not run:
                    return
                run["error"] = str(exc)
                run["state"] = AIRunState.FAILED
                run["finished_at"] = _now_iso()
                self._stats["failed"] += 1

    def _trim_old_runs_locked(self):
        if len(self._runs) <= self._max_runs:
            return
        ordered = sorted(
            self._runs.values(),
            key=lambda r: r.get("submitted_at") or "",
        )
        drop_count = len(self._runs) - self._max_runs
        for run in ordered[:drop_count]:
            self._runs.pop(run["run_id"], None)
            self._stats["dropped_trimmed"] += 1


def _age_seconds(ts_iso):
    if not ts_iso:
        return 0.0
    try:
        then = datetime.fromisoformat(str(ts_iso).replace("Z", "+00:00"))
    except Exception:
        return 0.0
    return max(0.0, (datetime.now(timezone.utc) - then).total_seconds())



import uuid
import random
from hiveos.db import get_db
from hiveos.models.chunk import Chunk
from hiveos.agent_interface import AgentInterface
from hiveos.utilities.log_event import log_chunk_event

# Runtime registry here instead of in agent_runtime.py
instances = []

def register(agent):
    instances.append(agent)

def get_agent_by_name(name):
    return next((a for a in instances if a.name == name), None)

class SimAgent(AgentInterface):
    _name_counter = 1

    def __init__(self, agent_id=None, name=None, battery=100, zone=None, capabilities='[]', residency=None):
        self.agent_id = agent_id or str(uuid.uuid4())
        print(f"[DEBUG] SimAgent.__init__: initializing agent {self.agent_id[:8]}")
        self.status = 'idle'
        self.battery = battery
        self.zone = zone
        self.capabilities = capabilities
        self.zone_residency_until = residency

        if name:
            self.name = name
            print(f"[DEBUG] Assigned name directly: {self.name}")
        else:
            try:
                from hiveos.agent_runtime import agent_objects
                if self.agent_id in agent_objects:
                    self.name = agent_objects[self.agent_id].name
                else:
                    self.name = f"Agent-{SimAgent._name_counter:03d}"
                    print(f"[DEBUG] Assigned auto-name: {self.name}")
                    SimAgent._name_counter += 1
            except ImportError:
                self.name = f"Agent-{SimAgent._name_counter:03d}"
                print(f"[DEBUG] Assigned auto-name: {self.name}")
                SimAgent._name_counter += 1

        self.current_chunk = None
        self.chunk_progress = 0
        self.required_ticks = 0
        self.cooldown_ticks = 0

        self.register()
        from hiveos.agent_runtime import register
        register(self)

    def register(self):
        with get_db() as conn:
            conn.execute("""
                INSERT OR REPLACE INTO agents (agent_id, status, battery, zone, capabilities, zone_residency_until)
                VALUES (?, ?, ?, ?, ?, ?)
            """, (self.agent_id, self.status, self.battery, self.zone, self.capabilities, self.zone_residency_until))

    def get_status(self):
        return self.status

    def report_capabilities(self):
        return self.capabilities

    def execute_chunk(self, chunk_id):
        print(f"[{self.name}] Executing chunk {chunk_id}...")
        self.status = 'working'
        self.current_chunk = chunk_id
        self.chunk_progress = 0
        self.required_ticks = random.randint(3, 7)
        self.register()

    def report_completion(self, chunk_id):
        print(f"[{self.name}] Completed chunk {chunk_id}.")
        self.status = 'idle'
        self.current_chunk = None
        self.chunk_progress = 0
        self.required_ticks = 0
        Chunk.complete(chunk_id)
        self.register()

    def enter_cooldown(self, reason=""):
        print(f"[{self.name}] Entering cooldown: {reason}")
        self.status = 'cooldown'
        self.cooldown_ticks = random.randint(3, 6)
        self.register()

    def is_available(self):
        return self.status == 'idle'

    def tick(self):
        if self.status == 'cooldown':
            self.cooldown_ticks -= 1
            if self.cooldown_ticks <= 0:
                print(f"[{self.name}] Cooldown complete.")
                self.status = 'idle'
                self.register()
            return

        if self.status == 'recharging':
            self.cooldown_ticks -= 1
            if self.cooldown_ticks <= 0:
                print(f"[{self.name}] Recharged.")
                self.status = 'idle'
                self.battery = 100
                self.register()
            return

        if self.status == 'working':
            self.chunk_progress += 1
            self.battery = max(0, self.battery - 1)

            if self.battery == 0:
                print(f"[{self.name}] Battery depleted. Dropping chunk {self.current_chunk}.")
                task_id = Chunk.get_task_id(self.current_chunk)
                log_chunk_event(self.current_chunk, task_id, self.agent_id, "fallback", reason="battery depleted")
                Chunk.reset(self.current_chunk)
                self.status = 'recharging'
                self.cooldown_ticks = random.randint(5, 10)
                self.current_chunk = None
                self.chunk_progress = 0
                self.required_ticks = 0
                self.register()
                return

            if self.chunk_progress >= self.required_ticks:
                self.report_completion(self.current_chunk)

    def health_check(self):
        return {'battery': self.battery, 'status': self.status}

    def score(self, task_context=None):
        """
        TODO: Return a dynamic score representing this agent's fitness
        to be assigned work. Could include battery, history, latency,
        or custom telemetry when real data is available.
        """
        return 1.0  # neutral score for now



from flask import request

from hiveos.config import ENFORCE_USER_CAPABILITIES, USER_CAPABILITIES_HEADER
from hiveos.intelligence.tracing import build_trace_event, emit_trace, now_iso
from hiveos.security.authorization import authorize_action, parse_capabilities
from hiveos.session_manager import session_manager


def build_action_run_id(action, user_id=None, session_id=None):
    resolved_session = session_id if session_id is not None else request.headers.get("X-Session-ID")
    resolved_user = user_id if user_id is not None else request.headers.get("X-User-ID")
    return f"policy:{action}:{resolved_session or 'none'}:{resolved_user or 'none'}"


def emit_action_applied(action, resource_type, resource_id=None, user_id=None, session_id=None, payload=None):
    event_payload = {"action": action, "resource_type": resource_type, "resource_id": resource_id}
    if payload:
        event_payload.update(payload)
    emit_trace(
        build_trace_event(
            event_type="action_applied",
            run_id=build_action_run_id(action, user_id=user_id, session_id=session_id),
            outcome="success",
            payload=event_payload,
            task_id=(resource_id if resource_type == "task" else None),
            zone_id=(resource_id if resource_type == "zone" else None),
        )
    )


def emit_action_failed(action, resource_type, reason, resource_id=None, user_id=None, session_id=None, payload=None):
    event_payload = {
        "action": action,
        "resource_type": resource_type,
        "resource_id": resource_id,
        "reason": str(reason),
    }
    if payload:
        event_payload.update(payload)
    emit_trace(
        build_trace_event(
            event_type="action_failed",
            run_id=build_action_run_id(action, user_id=user_id, session_id=session_id),
            outcome="failed",
            payload=event_payload,
            task_id=(resource_id if resource_type == "task" else None),
            zone_id=(resource_id if resource_type == "zone" else None),
        )
    )


def get_core_from_request():
    session_id = request.headers.get("X-Session-ID")
    user_id = request.headers.get("X-User-ID")

    if not session_id or not user_id:
        raise PermissionError("Missing session or user ID")

    session = session_manager.get_session_record(session_id)
    if not session or session.get("user_id") != user_id:
        raise PermissionError("Unauthorized or missing session")

    return session["core"]


def authorize_core_action(action, resource_type, resource_id=None):
    session_id = request.headers.get("X-Session-ID")
    user_id = request.headers.get("X-User-ID")
    caps_raw = request.headers.get(USER_CAPABILITIES_HEADER)
    granted_caps = parse_capabilities(caps_raw)
    run_id = build_action_run_id(action, user_id=user_id, session_id=session_id)

    emit_trace(
        build_trace_event(
            event_type="action_requested",
            run_id=run_id,
            outcome="success",
            payload={
                "action": action,
                "resource_type": resource_type,
                "resource_id": resource_id,
                "actor": {"user_id": user_id, "session_id": session_id, "capabilities": granted_caps},
                "requested_at": now_iso(),
            },
            task_id=(resource_id if resource_type == "task" else None),
            zone_id=(resource_id if resource_type == "zone" else None),
        )
    )

    core = get_core_from_request()
    decision = authorize_action(action, granted_caps, enforce=ENFORCE_USER_CAPABILITIES)
    if not decision.allowed:
        emit_trace(
            build_trace_event(
                event_type="action_denied",
                run_id=run_id,
                outcome="denied",
                payload={
                    "action": action,
                    "resource_type": resource_type,
                    "resource_id": resource_id,
                    "reason": decision.reason,
                    "required_capabilities": list(decision.required_capabilities),
                    "granted_capabilities": list(decision.granted_capabilities),
                },
                task_id=(resource_id if resource_type == "task" else None),
                zone_id=(resource_id if resource_type == "zone" else None),
            )
        )
        raise PermissionError(f"Unauthorized action: {decision.reason}")

    emit_trace(
        build_trace_event(
            event_type="action_authorized",
            run_id=run_id,
            outcome="success",
            payload={
                "action": action,
                "resource_type": resource_type,
                "resource_id": resource_id,
                "required_capabilities": list(decision.required_capabilities),
                "granted_capabilities": list(decision.granted_capabilities),
                "reason": decision.reason,
            },
            task_id=(resource_id if resource_type == "task" else None),
            zone_id=(resource_id if resource_type == "zone" else None),
        )
    )
    return core


def authorize_user_action(action, resource_type, user_id, session_id=None, resource_id=None):
    caps_raw = request.headers.get(USER_CAPABILITIES_HEADER)
    granted_caps = parse_capabilities(caps_raw)
    run_id = build_action_run_id(action, user_id=user_id, session_id=session_id)

    emit_trace(
        build_trace_event(
            event_type="action_requested",
            run_id=run_id,
            outcome="success",
            payload={
                "action": action,
                "resource_type": resource_type,
                "resource_id": resource_id,
                "actor": {"user_id": user_id, "session_id": session_id, "capabilities": granted_caps},
                "requested_at": now_iso(),
            },
            task_id=(resource_id if resource_type == "task" else None),
            zone_id=(resource_id if resource_type == "zone" else None),
        )
    )

    decision = authorize_action(action, granted_caps, enforce=ENFORCE_USER_CAPABILITIES)
    if not decision.allowed:
        emit_trace(
            build_trace_event(
                event_type="action_denied",
                run_id=run_id,
                outcome="denied",
                payload={
                    "action": action,
                    "resource_type": resource_type,
                    "resource_id": resource_id,
                    "reason": decision.reason,
                    "required_capabilities": list(decision.required_capabilities),
                    "granted_capabilities": list(decision.granted_capabilities),
                },
                task_id=(resource_id if resource_type == "task" else None),
                zone_id=(resource_id if resource_type == "zone" else None),
            )
        )
        raise PermissionError(f"Unauthorized action: {decision.reason}")

    emit_trace(
        build_trace_event(
            event_type="action_authorized",
            run_id=run_id,
            outcome="success",
            payload={
                "action": action,
                "resource_type": resource_type,
                "resource_id": resource_id,
                "required_capabilities": list(decision.required_capabilities),
                "granted_capabilities": list(decision.granted_capabilities),
                "reason": decision.reason,
            },
            task_id=(resource_id if resource_type == "task" else None),
            zone_id=(resource_id if resource_type == "zone" else None),
        )
    )



import uuid
import json
from multiprocessing import Queue
import queue
import time
from hiveos.agent_interface import AgentInterface
from hiveos.runtime.states import AgentStatus

class SimTerminalAgent(AgentInterface):
    def __init__(self, agent_id=None, name=None, capabilities=None):
        super().__init__()
        self.agent_id = agent_id or f"sim-terminal-{str(uuid.uuid4())[:8]}"
        self.name = str(name or f"SimTerm-{str(uuid.uuid4())[:4]}")
        self.zone = None
        self.status = AgentStatus.IDLE
        self.capabilities = ["terminal", "display"]
        self._available = True
        self.current_chunk = None
        self.chunk_progress = 0
        self.required_ticks = 0
        self.waiting_on_input = False

        self.input_queue = Queue()
        self.latest_output = ""
        self.latest_input = None
        self.wait_started_mono = None
        self.wait_timeout_seconds = None
        self._wait_token = None
        self._core = None

    def bind_core(self, core):
        self._core = core

    def get_status(self):
        return self.status

    def report_capabilities(self):
        return self.capabilities

    def execute_chunk(self, chunk, payload=None):
        # Backwards compatibility: wrap legacy string payload in dict
        if isinstance(payload, str):
            payload = {"intent": "render_text", "parameters": {"text": payload}}

        intent = payload.get("intent", "")
        self.current_intent = intent
        params = payload.get("parameters", {})

        if intent == "render_text":
            text = params.get("text", "")
            self.latest_output = text                     
            self.execution_started = True
            self.ready_to_execute = False
        
        elif intent == "wait_for_input":
            if not self.waiting_on_input:
                # Trigger prompt as soon as the chunk is assigned and ready
                prompt = params.get("prompt", "Enter input:")
                self.latest_output = prompt
                self.waiting_on_input = True
                self.wait_started_mono = time.monotonic()
                self.wait_timeout_seconds = self._resolve_wait_timeout_seconds(chunk, params)
                self._wait_token = f"{chunk.chunk_id}:{time.time_ns()}"
                wait_token = self._wait_token

                def wait_and_publish():
                    while True:
                        if not self.waiting_on_input or self.current_chunk != chunk.chunk_id:
                            break
                        if self._wait_token != wait_token:
                            break
                        try:
                            user_input = self.input_queue.get(timeout=0.2)
                        except queue.Empty:
                            continue
                        self.latest_input = user_input
                        if self.message_bus is not None:
                            self.message_bus.publish("terminal_input", user_input)
                        self.waiting_on_input = False
                        self.wait_started_mono = None
                        self.wait_timeout_seconds = None
                        self.report_completion(chunk)
                        break

                import threading
                threading.Thread(target=wait_and_publish, daemon=True).start()
        elif intent == "terminal_action":
            self._execute_terminal_action(chunk, payload)
        else:
            self.latest_output = f"Unknown intent: {intent}"

    def tick(self):
        super().tick()
        if not self.waiting_on_input or self.wait_started_mono is None:
            return
        if not self.wait_timeout_seconds or self.wait_timeout_seconds <= 0:
            return
        if (time.monotonic() - self.wait_started_mono) < self.wait_timeout_seconds:
            return

        self.waiting_on_input = False
        self.wait_started_mono = None
        self.wait_timeout_seconds = None
        timeout_msg = "terminal_input_timeout"
        self.latest_input = timeout_msg
        if self.message_bus is not None:
            self.message_bus.publish("terminal_input", timeout_msg)
        if self.chunk_obj is not None:
            self.report_completion(self.chunk_obj)
    
    def health_check(self):
        return True

    def enter_cooldown(self, reason=""):
        self.status = AgentStatus.COOLDOWN
        self.latest_output = f"[COOLDOWN] {reason}"

    def is_available(self):
        return self._available and self.status == AgentStatus.IDLE
        
    def receive_input(self, user_input: str):
        if self.waiting_on_input:
            self.input_queue.put(user_input)

    def get_summary(self):
            return super().get_summary() | {
                "latest_output": self.latest_output,
                "latest_input": self.latest_input,
                "current_intent": getattr(self, "current_intent", None),
                "waiting_on_input": self.waiting_on_input,
                "wait_timeout_seconds": self.wait_timeout_seconds,
            }

    def _resolve_wait_timeout_seconds(self, chunk, params):
        raw = params.get("timeout_seconds")
        if raw is not None:
            try:
                return max(1.0, float(raw))
            except Exception:
                pass

        if chunk.timing == "seconds":
            return max(1.0, float(chunk.ttl))
        if chunk.timing == "ms":
            return max(0.25, float(chunk.ttl) / 1000.0)
        # Tick-based fallback: 1 tick ~= 0.1s with floor/ceiling guard.
        return max(2.0, min(300.0, float(chunk.ttl) * 0.1))

    def _execute_terminal_action(self, chunk, payload):
        if self._core is None:
            self.latest_output = "terminal_action_failed:missing_core_binding"
            self._mark_chunk_failed(chunk)
            return
        try:
            tool_id = str(payload.get("tool_id") or "")
            action = str(payload.get("action") or "")
            args = dict(payload.get("args") or {})
            trace_context = dict(payload.get("trace_context") or {})
            trace_context.setdefault("agent_id", self.agent_id)
            trace_context.setdefault("task_id", chunk.task_id)
            trace_context.setdefault("chunk_id", chunk.chunk_id)
            trace_context.setdefault("zone_id", self.zone)

            if tool_id == "terminal.session" and action == "open":
                result = self._core.terminal_service.open_target(
                    {
                        "target_id": args.get("target_id"),
                        "target_type": args.get("target_type"),
                        "session_id": self._core.session_id,
                        "agent_id": self.agent_id,
                        "transport": args.get("transport"),
                        "meta": args.get("meta"),
                    },
                    trace_context=trace_context,
                )
            elif tool_id == "terminal.transform" and action == "apply":
                result = self._core.terminal_service.set_transform(
                    target_id=str(args.get("target_id") or ""),
                    module=str(args.get("module") or ""),
                    config=args.get("config") or {},
                    trace_context=trace_context,
                )
            elif tool_id == "terminal.exec" and action == "run":
                result = self._core.terminal_service.run_command(
                    target_id=str(args.get("target_id") or ""),
                    command=str(args.get("command") or ""),
                    cwd=args.get("cwd"),
                    env=args.get("env"),
                    timeout_sec=int(args.get("timeout_sec") or 30),
                    trace_context=trace_context,
                )
            else:
                raise ValueError(f"unsupported_terminal_action:{tool_id}:{action}")

            self.latest_output = f"terminal_action_applied:{json.dumps({'tool_id': tool_id, 'action': action})}"
            self.report_completion(chunk)
            return result
        except Exception as exc:
            self.latest_output = f"terminal_action_failed:{exc}"
            self._mark_chunk_failed(chunk)
            return None

    def _mark_chunk_failed(self, chunk):
        self.has_executed = True
        self.status = AgentStatus.IDLE
        self.current_chunk = None
        self.chunk_progress = 0
        self.required_ticks = 0
        self.execution_started = False
        self.ready_to_execute = False
        if chunk is not None:
            chunk.mark_stale()



import json
import ctypes
import os
import shutil
import subprocess
import sys
import tempfile
import time
import uuid
from collections import deque
from datetime import datetime, timezone
from threading import RLock


def _now_iso():
    return datetime.now(timezone.utc).isoformat()


class SandboxRuntimeState:
    QUEUED = "queued"
    RUNNING = "running"
    COMPLETED = "completed"
    CANCELED = "canceled"
    TIMEOUT = "timeout"
    FAILED = "failed"


class SandboxRuntimeManager:
    """
    Sandboxed evolution runtime manager with subprocess isolation.
    Applies platform-specific hard limits where available.
    """

    def __init__(self, max_runs=500, max_runtime_seconds=20, max_memory_mb=512, max_cpu_seconds=10):
        self._max_runs = max_runs
        self._max_runtime_seconds = max_runtime_seconds
        self._max_memory_mb = max_memory_mb
        self._max_cpu_seconds = max_cpu_seconds
        self._lock = RLock()
        self._queue = deque()
        self._runs = {}
        self._active = None  # {"sandbox_run_id", "proc", "started_monotonic", "timeout_s", "cwd", "job_handle"}
        self._events = deque(maxlen=2000)
        self._stats = {
            "started": 0,
            "canceled": 0,
            "completed": 0,
            "failed": 0,
            "timeouts": 0,
            "isolation_applied": 0,
            "isolation_degraded": 0,
        }

    def start_run(self, run_spec):
        if not isinstance(run_spec, dict):
            raise ValueError("run_spec must be an object")
        constraints = self._normalize_constraints(run_spec.get("constraints") or {})
        timeout_s = self._resolve_timeout(constraints)

        sandbox_run_id = f"sbox-{uuid.uuid4().hex[:10]}"
        now = _now_iso()
        run = {
            "sandbox_run_id": sandbox_run_id,
            "state": SandboxRuntimeState.QUEUED,
            "created_at": now,
            "started_at": None,
            "finished_at": None,
            "objective": run_spec.get("objective", ""),
            "task_id": run_spec.get("task_id"),
            "chunk_id": run_spec.get("chunk_id"),
            "agent_id": run_spec.get("agent_id"),
            "constraints": dict(constraints),
            "timeout_seconds": timeout_s,
            "snapshot_ref": run_spec.get("snapshot_ref"),
            "result": None,
            "error": None,
        }

        with self._lock:
            self._runs[sandbox_run_id] = run
            self._queue.append(sandbox_run_id)
            self._stats["started"] += 1
            self._trim_old_runs_locked()
        return sandbox_run_id

    def get_result(self, sandbox_run_id):
        with self._lock:
            run = self._runs.get(sandbox_run_id)
            return dict(run) if run else None

    def cancel(self, sandbox_run_id):
        with self._lock:
            run = self._runs.get(sandbox_run_id)
            if not run:
                return False
            if run["state"] in {
                SandboxRuntimeState.COMPLETED,
                SandboxRuntimeState.CANCELED,
                SandboxRuntimeState.TIMEOUT,
                SandboxRuntimeState.FAILED,
            }:
                return False
            if self._active and self._active["sandbox_run_id"] == sandbox_run_id:
                proc = self._active["proc"]
                if proc.poll() is None:
                    proc.terminate()
            run["state"] = SandboxRuntimeState.CANCELED
            run["finished_at"] = _now_iso()
            self._stats["canceled"] += 1
            self._events.append({"type": "canceled", "sandbox_run_id": sandbox_run_id, "timestamp": _now_iso()})
            return True

    def process_next(self):
        """
        Cooperative non-blocking step:
        - polls active subprocess-backed sandbox run
        - enforces run timeout
        - launches queued run if slot is available
        """
        self._poll_active()
        self._launch_next_if_idle()

    def drain_events(self, max_events=100):
        with self._lock:
            events = []
            for _ in range(min(max_events, len(self._events))):
                events.append(self._events.popleft())
            return events

    def get_stats(self):
        with self._lock:
            queued = sum(1 for r in self._runs.values() if r["state"] == SandboxRuntimeState.QUEUED)
            running = sum(1 for r in self._runs.values() if r["state"] == SandboxRuntimeState.RUNNING)
            timeouts = sum(1 for r in self._runs.values() if r["state"] == SandboxRuntimeState.TIMEOUT)
            return {
                "total_runs": len(self._runs),
                "queued_runs": queued,
                "running_runs": running,
                "timeout_runs": timeouts,
                "active_process": bool(self._active),
                **self._stats,
            }

    def _trim_old_runs_locked(self):
        if len(self._runs) <= self._max_runs:
            return
        ordered = sorted(self._runs.values(), key=lambda r: r.get("created_at") or "")
        drop_count = len(self._runs) - self._max_runs
        for run in ordered[:drop_count]:
            self._runs.pop(run["sandbox_run_id"], None)

    def _resolve_timeout(self, constraints):
        requested = constraints.get("max_runtime_seconds", self._max_runtime_seconds)
        try:
            timeout_s = int(requested)
        except Exception as exc:
            raise ValueError(f"invalid max_runtime_seconds: {exc}") from exc
        if timeout_s < 1:
            timeout_s = 1
        if timeout_s > self._max_runtime_seconds:
            timeout_s = self._max_runtime_seconds
        return timeout_s

    def _normalize_constraints(self, constraints):
        if not isinstance(constraints, dict):
            raise ValueError("constraints must be an object")

        allowed = {
            "max_runtime_seconds",
            "max_memory_mb",
            "max_cpu_seconds",
            "max_open_files",
            "no_network",
            "read_only_fs",
            "simulate_work_ms",
        }
        unknown = [k for k in constraints.keys() if k not in allowed]
        if unknown:
            raise ValueError(f"unsupported constraints: {unknown}")

        no_network = constraints.get("no_network", True)
        read_only_fs = constraints.get("read_only_fs", True)
        if no_network is not True:
            raise ValueError("no_network must be true in sandbox policy")
        if read_only_fs is not True:
            raise ValueError("read_only_fs must be true in sandbox policy")

        memory = constraints.get("max_memory_mb", self._max_memory_mb)
        cpu = constraints.get("max_cpu_seconds", self._max_cpu_seconds)
        max_open_files = constraints.get("max_open_files", 64)
        runtime = constraints.get("max_runtime_seconds", self._max_runtime_seconds)
        simulate_ms = constraints.get("simulate_work_ms", 0)

        try:
            memory = int(memory)
            cpu = int(cpu)
            max_open_files = int(max_open_files)
            runtime = int(runtime)
            simulate_ms = int(simulate_ms)
        except Exception as exc:
            raise ValueError(f"invalid numeric constraint: {exc}") from exc

        if memory < 64:
            memory = 64
        if memory > self._max_memory_mb:
            memory = self._max_memory_mb
        if cpu < 1:
            cpu = 1
        if cpu > self._max_cpu_seconds:
            cpu = self._max_cpu_seconds
        if runtime < 1:
            runtime = 1
        if runtime > self._max_runtime_seconds:
            runtime = self._max_runtime_seconds
        if max_open_files < 16:
            max_open_files = 16
        if max_open_files > 256:
            max_open_files = 256
        if simulate_ms < 0:
            simulate_ms = 0

        return {
            "max_runtime_seconds": runtime,
            "max_memory_mb": memory,
            "max_cpu_seconds": cpu,
            "max_open_files": max_open_files,
            "no_network": True,
            "read_only_fs": True,
            "simulate_work_ms": simulate_ms,
        }

    def _launch_next_if_idle(self):
        with self._lock:
            if self._active is not None:
                return
            sandbox_run_id = self._queue.popleft() if self._queue else None
            if not sandbox_run_id:
                return
            run = self._runs.get(sandbox_run_id)
            if not run or run["state"] != SandboxRuntimeState.QUEUED:
                return

            payload = {
                "sandbox_run_id": sandbox_run_id,
                "objective": run.get("objective"),
                "constraints": run.get("constraints", {}),
                "snapshot_ref": run.get("snapshot_ref"),
                "task_id": run.get("task_id"),
                "chunk_id": run.get("chunk_id"),
                "agent_id": run.get("agent_id"),
            }
            sandbox_cwd = tempfile.mkdtemp(prefix=f"hive_sbox_{sandbox_run_id}_")
            env = _build_worker_env()
            env["PYTHONNOUSERSITE"] = "1"
            env["HIVE_SANDBOX"] = "1"
            env["PYTHONIOENCODING"] = "utf-8"
            env["PYTHONUNBUFFERED"] = "1"
            worker_path = os.path.join(os.path.dirname(__file__), "sandbox_worker.py")
            popen_kwargs = {
                "stdin": subprocess.PIPE,
                "stdout": subprocess.PIPE,
                "stderr": subprocess.PIPE,
                "text": True,
                "cwd": sandbox_cwd,
                "env": env,
            }
            if os.name == "posix":
                popen_kwargs["preexec_fn"] = self._build_posix_preexec(run.get("constraints", {}))
            if os.name == "nt":
                creationflags = 0
                creationflags |= getattr(subprocess, "CREATE_NEW_PROCESS_GROUP", 0)
                creationflags |= getattr(subprocess, "CREATE_NO_WINDOW", 0)
                creationflags |= 0x01000000  # CREATE_BREAKAWAY_FROM_JOB
                popen_kwargs["creationflags"] = creationflags
            proc = subprocess.Popen(
                [sys.executable, "-I", worker_path],
                **popen_kwargs,
            )
            proc.stdin.write(json.dumps(payload, separators=(",", ":")))
            proc.stdin.close()
            job_handle = None
            if os.name == "nt":
                job_handle = self._assign_windows_job_limits(proc, run.get("constraints", {}))
                if job_handle:
                    self._stats["isolation_applied"] += 1
                    self._events.append(
                        {
                            "type": "isolation_applied",
                            "sandbox_run_id": sandbox_run_id,
                            "timestamp": _now_iso(),
                            "backend": "windows_job_object",
                        }
                    )
                else:
                    self._stats["isolation_degraded"] += 1
                    self._events.append(
                        {
                            "type": "isolation_degraded",
                            "sandbox_run_id": sandbox_run_id,
                            "timestamp": _now_iso(),
                            "backend": "windows_job_object",
                        }
                    )
            elif os.name == "posix":
                self._stats["isolation_applied"] += 1
                self._events.append(
                    {
                        "type": "isolation_applied",
                        "sandbox_run_id": sandbox_run_id,
                        "timestamp": _now_iso(),
                        "backend": "posix_setrlimit",
                    }
                )
            run["state"] = SandboxRuntimeState.RUNNING
            run["started_at"] = _now_iso()
            self._active = {
                "sandbox_run_id": sandbox_run_id,
                "proc": proc,
                "started_monotonic": time.monotonic(),
                "timeout_s": int(run.get("timeout_seconds", self._max_runtime_seconds)),
                "cwd": sandbox_cwd,
                "job_handle": job_handle,
            }
            self._events.append({"type": "started", "sandbox_run_id": sandbox_run_id, "timestamp": _now_iso()})

    def _poll_active(self):
        with self._lock:
            active = dict(self._active) if self._active else None
        if not active:
            return

        sandbox_run_id = active["sandbox_run_id"]
        proc = active["proc"]
        run = self.get_result(sandbox_run_id)
        if not run:
            with self._lock:
                self._active = None
            return

        if proc.poll() is None:
            elapsed = time.monotonic() - active["started_monotonic"]
            if elapsed > active["timeout_s"]:
                proc.kill()
                proc.communicate(timeout=0.2)
                with self._lock:
                    live = self._runs.get(sandbox_run_id)
                    if live and live["state"] == SandboxRuntimeState.RUNNING:
                        live["state"] = SandboxRuntimeState.TIMEOUT
                        live["finished_at"] = _now_iso()
                        live["error"] = f"timeout>{active['timeout_s']}s"
                        self._stats["timeouts"] += 1
                        self._events.append(
                            {"type": "timeout", "sandbox_run_id": sandbox_run_id, "timestamp": _now_iso()}
                        )
                    self._active = None
                self._close_windows_job_handle(active.get("job_handle"))
                self._cleanup_sandbox_dir(active.get("cwd"))
            return

        stdout, stderr = proc.communicate(timeout=0.2)
        if proc.returncode != 0:
            with self._lock:
                live = self._runs.get(sandbox_run_id)
                if live and live["state"] == SandboxRuntimeState.RUNNING:
                    live["state"] = SandboxRuntimeState.FAILED
                    live["finished_at"] = _now_iso()
                    live["error"] = (stderr or f"worker_exit_{proc.returncode}").strip()[:1000]
                    self._stats["failed"] += 1
                    self._events.append({"type": "failed", "sandbox_run_id": sandbox_run_id, "timestamp": _now_iso()})
                self._active = None
            self._close_windows_job_handle(active.get("job_handle"))
            self._cleanup_sandbox_dir(active.get("cwd"))
            return

        try:
            raw_result = json.loads((stdout or "").strip() or "{}")
        except Exception:
            raw_result = {"artifact_only": True, "message": "worker returned non-json output"}

        sanitized = self._sanitize_worker_result(raw_result)
        with self._lock:
            live = self._runs.get(sandbox_run_id)
            if live and live["state"] == SandboxRuntimeState.RUNNING:
                live["state"] = SandboxRuntimeState.COMPLETED
                live["finished_at"] = _now_iso()
                live["result"] = sanitized
                self._stats["completed"] += 1
                self._events.append({"type": "completed", "sandbox_run_id": sandbox_run_id, "timestamp": _now_iso()})
            self._active = None
        self._close_windows_job_handle(active.get("job_handle"))
        self._cleanup_sandbox_dir(active.get("cwd"))

    def _sanitize_worker_result(self, raw):
        if not isinstance(raw, dict):
            raw = {}
        allowed_keys = {"artifact_only", "proposal_candidates", "metrics", "traces", "message", "meta"}
        sanitized = {k: raw.get(k) for k in allowed_keys if k in raw}
        sanitized["artifact_only"] = True
        if "proposal_candidates" not in sanitized or not isinstance(sanitized["proposal_candidates"], list):
            sanitized["proposal_candidates"] = []
        if "metrics" not in sanitized or not isinstance(sanitized["metrics"], dict):
            sanitized["metrics"] = {}
        if "traces" not in sanitized or not isinstance(sanitized["traces"], list):
            sanitized["traces"] = []
        return sanitized

    def _cleanup_sandbox_dir(self, path):
        if not path:
            return
        try:
            shutil.rmtree(path, ignore_errors=True)
        except Exception:
            pass

    def _build_posix_preexec(self, constraints):
        def _preexec():
            import os as _os
            import resource as _resource

            mem_bytes = int(constraints.get("max_memory_mb", self._max_memory_mb)) * 1024 * 1024
            cpu_s = int(constraints.get("max_cpu_seconds", self._max_cpu_seconds))
            nofile = int(constraints.get("max_open_files", 64))

            _os.setsid()
            if hasattr(_resource, "RLIMIT_AS"):
                _resource.setrlimit(_resource.RLIMIT_AS, (mem_bytes, mem_bytes))
            if hasattr(_resource, "RLIMIT_CPU"):
                _resource.setrlimit(_resource.RLIMIT_CPU, (cpu_s, cpu_s))
            if hasattr(_resource, "RLIMIT_NOFILE"):
                _resource.setrlimit(_resource.RLIMIT_NOFILE, (nofile, nofile))
            if hasattr(_resource, "RLIMIT_NPROC"):
                _resource.setrlimit(_resource.RLIMIT_NPROC, (1, 1))

        return _preexec

    def _assign_windows_job_limits(self, proc, constraints):
        try:
            kernel32 = ctypes.windll.kernel32
        except Exception:
            return None

        class JOBOBJECT_BASIC_LIMIT_INFORMATION(ctypes.Structure):
            _fields_ = [
                ("PerProcessUserTimeLimit", ctypes.c_longlong),
                ("PerJobUserTimeLimit", ctypes.c_longlong),
                ("LimitFlags", ctypes.c_uint32),
                ("MinimumWorkingSetSize", ctypes.c_size_t),
                ("MaximumWorkingSetSize", ctypes.c_size_t),
                ("ActiveProcessLimit", ctypes.c_uint32),
                ("Affinity", ctypes.c_size_t),
                ("PriorityClass", ctypes.c_uint32),
                ("SchedulingClass", ctypes.c_uint32),
            ]

        class IO_COUNTERS(ctypes.Structure):
            _fields_ = [
                ("ReadOperationCount", ctypes.c_uint64),
                ("WriteOperationCount", ctypes.c_uint64),
                ("OtherOperationCount", ctypes.c_uint64),
                ("ReadTransferCount", ctypes.c_uint64),
                ("WriteTransferCount", ctypes.c_uint64),
                ("OtherTransferCount", ctypes.c_uint64),
            ]

        class JOBOBJECT_EXTENDED_LIMIT_INFORMATION(ctypes.Structure):
            _fields_ = [
                ("BasicLimitInformation", JOBOBJECT_BASIC_LIMIT_INFORMATION),
                ("IoInfo", IO_COUNTERS),
                ("ProcessMemoryLimit", ctypes.c_size_t),
                ("JobMemoryLimit", ctypes.c_size_t),
                ("PeakProcessMemoryUsed", ctypes.c_size_t),
                ("PeakJobMemoryUsed", ctypes.c_size_t),
            ]

        JOB_OBJECT_LIMIT_PROCESS_TIME = 0x00000002
        JOB_OBJECT_LIMIT_ACTIVE_PROCESS = 0x00000008
        JOB_OBJECT_LIMIT_PROCESS_MEMORY = 0x00000100
        JOB_OBJECT_LIMIT_KILL_ON_JOB_CLOSE = 0x00002000
        JOB_OBJECT_EXTENDED_LIMIT_INFORMATION_CLASS = 9

        job = kernel32.CreateJobObjectW(None, None)
        if not job:
            return None

        info = JOBOBJECT_EXTENDED_LIMIT_INFORMATION()
        info.BasicLimitInformation.LimitFlags = (
            JOB_OBJECT_LIMIT_PROCESS_TIME
            | JOB_OBJECT_LIMIT_ACTIVE_PROCESS
            | JOB_OBJECT_LIMIT_PROCESS_MEMORY
            | JOB_OBJECT_LIMIT_KILL_ON_JOB_CLOSE
        )
        cpu_s = int(constraints.get("max_cpu_seconds", self._max_cpu_seconds))
        info.BasicLimitInformation.PerProcessUserTimeLimit = int(cpu_s * 10_000_000)
        info.BasicLimitInformation.ActiveProcessLimit = 1
        info.ProcessMemoryLimit = int(constraints.get("max_memory_mb", self._max_memory_mb)) * 1024 * 1024

        ok = kernel32.SetInformationJobObject(
            job,
            JOB_OBJECT_EXTENDED_LIMIT_INFORMATION_CLASS,
            ctypes.byref(info),
            ctypes.sizeof(info),
        )
        if not ok:
            kernel32.CloseHandle(job)
            return None

        ok = kernel32.AssignProcessToJobObject(job, ctypes.c_void_p(int(proc._handle)))
        if not ok:
            kernel32.CloseHandle(job)
            return None
        return job

    def _close_windows_job_handle(self, job_handle):
        if not job_handle:
            return
        try:
            ctypes.windll.kernel32.CloseHandle(job_handle)
        except Exception:
            pass


def _build_worker_env():
    # Build a minimal environment for the sandbox worker to reduce ambient authority.
    allowed_keys = {
        "SYSTEMROOT",
        "WINDIR",
        "PATH",
        "PATHEXT",
        "COMSPEC",
        "TMP",
        "TEMP",
        "HOME",
        "USERPROFILE",
        "LANG",
        "LC_ALL",
    }
    src = os.environ
    env = {}
    for key in allowed_keys:
        value = src.get(key)
        if value is not None:
            env[key] = value
    return env



import json
import os
import socket
import subprocess
import sys
import time


def _install_network_guard():
    class _DeniedSocket(socket.socket):
        def connect(self, *args, **kwargs):  # pragma: no cover - defensive
            raise PermissionError("sandbox network access denied")

        def connect_ex(self, *args, **kwargs):  # pragma: no cover - defensive
            raise PermissionError("sandbox network access denied")

    socket.socket = _DeniedSocket
    socket.create_connection = _deny_callable("sandbox network access denied")
    socket.getaddrinfo = _deny_callable("sandbox dns resolution denied")


def _install_readonly_fs_guard():
    import builtins

    original_open = builtins.open

    def guarded_open(file, mode="r", *args, **kwargs):
        write_tokens = ("w", "a", "x", "+")
        if any(token in mode for token in write_tokens):
            raise PermissionError("sandbox filesystem write denied")
        return original_open(file, mode, *args, **kwargs)

    builtins.open = guarded_open
    if hasattr(os, "open"):
        original_os_open = os.open

        def guarded_os_open(path, flags, *args, **kwargs):
            write_flags = (
                getattr(os, "O_WRONLY", 0)
                | getattr(os, "O_RDWR", 0)
                | getattr(os, "O_CREAT", 0)
                | getattr(os, "O_TRUNC", 0)
                | getattr(os, "O_APPEND", 0)
            )
            if flags & write_flags:
                raise PermissionError("sandbox filesystem write denied")
            return original_os_open(path, flags, *args, **kwargs)

        os.open = guarded_os_open

    for name in (
        "remove",
        "unlink",
        "rename",
        "replace",
        "mkdir",
        "makedirs",
        "rmdir",
        "removedirs",
        "chmod",
        "chown",
    ):
        if hasattr(os, name):
            setattr(os, name, _deny_callable("sandbox filesystem mutation denied"))


def _install_subprocess_guard():
    if hasattr(subprocess, "Popen"):
        subprocess.Popen = _deny_callable("sandbox subprocess denied")
    for name in ("run", "call", "check_call", "check_output"):
        if hasattr(subprocess, name):
            setattr(subprocess, name, _deny_callable("sandbox subprocess denied"))
    if hasattr(os, "system"):
        os.system = _deny_callable("sandbox os.system denied")
    if hasattr(os, "popen"):
        os.popen = _deny_callable("sandbox os.popen denied")


def _deny_callable(message):
    def _denied(*_args, **_kwargs):  # pragma: no cover - defensive
        raise PermissionError(message)

    return _denied


def _apply_resource_limits(constraints):
    applied = {}
    unsupported = []
    try:
        import resource  # type: ignore
    except Exception:
        return {"applied": applied, "unsupported": ["resource_module_unavailable"]}

    max_memory_mb = int(constraints.get("max_memory_mb", 256))
    max_cpu_seconds = int(constraints.get("max_cpu_seconds", 5))
    try:
        resource.setrlimit(resource.RLIMIT_AS, (max_memory_mb * 1024 * 1024, max_memory_mb * 1024 * 1024))
        applied["RLIMIT_AS"] = max_memory_mb
    except Exception:
        unsupported.append("RLIMIT_AS")
    try:
        resource.setrlimit(resource.RLIMIT_CPU, (max_cpu_seconds, max_cpu_seconds))
        applied["RLIMIT_CPU"] = max_cpu_seconds
    except Exception:
        unsupported.append("RLIMIT_CPU")
    return {"applied": applied, "unsupported": unsupported}


def run_sandbox_spec(spec):
    """
    Process-isolated sandbox worker.
    Contract: emits artifact-only payload; never touches live core runtime.
    """
    started = time.monotonic()
    objective = str(spec.get("objective") or "").strip()
    constraints = spec.get("constraints") or {}
    if constraints.get("no_network", True):
        _install_network_guard()
    if constraints.get("read_only_fs", True):
        _install_readonly_fs_guard()
    _install_subprocess_guard()
    limit_info = _apply_resource_limits(constraints)

    simulate_ms = constraints.get("simulate_work_ms", 0)
    try:
        simulate_ms = int(simulate_ms)
    except Exception:
        simulate_ms = 0
    if simulate_ms > 0:
        time.sleep(min(simulate_ms / 1000.0, 0.5))

    proposal_candidates = []
    if objective:
        proposal_candidates.append(
            {
                "title": f"Candidate flow for: {objective[:80]}",
                "artifact_type": "task_draft",
                "confidence": 0.5,
                "notes": "sandbox_candidate_stub",
            }
        )

    elapsed_ms = int((time.monotonic() - started) * 1000)
    return {
        "artifact_only": True,
        "message": "sandbox worker completed",
        "proposal_candidates": proposal_candidates,
        "metrics": {"latency_ms": elapsed_ms, "candidate_count": len(proposal_candidates)},
        "traces": [{"step": "sandbox_worker", "objective": objective, "latency_ms": elapsed_ms}],
        "meta": {
            "worker": "hiveos.intelligence.sandbox_worker",
            "version": 2,
            "limits": limit_info,
            "sandbox_env": bool(os.environ.get("HIVE_SANDBOX")),
            "guards": {
                "network": bool(constraints.get("no_network", True)),
                "read_only_fs": bool(constraints.get("read_only_fs", True)),
                "subprocess_blocked": True,
            },
        },
    }


def main():
    raw = sys.stdin.read().strip()
    spec = json.loads(raw or "{}")
    result = run_sandbox_spec(spec)
    print(json.dumps(result, separators=(",", ":")))


if __name__ == "__main__":
    main()




def reset_runtime():
    from hiveos.sim_agent import SimAgent
    from hiveos.core_loop import core
    core.zones.clear()
    core._runtime_agents.clear()
    core.task_objects.clear()
    core.chunk_objects.clear()
    core.task_name_map.clear()
    core.tick_counter["count"] = 0
    SimAgent._name_counter = 1
    print("✅ Runtime state has been reset (agents, zones, counters).")



from copy import deepcopy


DEFAULT_POLICY = {
    "weights": {
        "validated": 8.0,
        "injected": 10.0,
        "created": 1.0,
        "rejected": -10.0,
        "task_draft": 2.0,
        "wrapper_draft": 0.5,
        "patch_draft": 0.5,
        "validation_error": -1.0,
        "confidence": 4.0,
    },
    "require_validated": True,
}


def evaluate_candidates(candidates, policy=None):
    policy_obj = _merge_policy(policy)
    ranked = []
    for idx, candidate in enumerate(candidates or []):
        score, breakdown = _score_candidate(candidate, policy_obj)
        ranked.append(
            {
                "rank": 0,
                "proposal_id": str(candidate.get("proposal_id") or f"candidate-{idx}"),
                "score": round(score, 6),
                "status": str(candidate.get("status") or "unknown"),
                "artifact_type": str(((candidate.get("artifact") or {}).get("artifact_type")) or "unknown"),
                "eligible": _is_eligible(candidate, policy_obj),
                "breakdown": breakdown,
            }
        )

    ranked.sort(key=lambda row: (-row["score"], row["proposal_id"]))
    for index, row in enumerate(ranked, start=1):
        row["rank"] = index
    return ranked


def select_winner(ranked_results, require_validated=True):
    ranked = list(ranked_results or [])
    if not ranked:
        return None
    if require_validated:
        for row in ranked:
            if row.get("eligible"):
                return row
        return None
    return ranked[0]


def _score_candidate(candidate, policy_obj):
    weights = policy_obj["weights"]
    score = 0.0
    breakdown = {}

    status = str(candidate.get("status") or "").lower()
    if status in weights:
        score += float(weights[status])
        breakdown[f"status:{status}"] = float(weights[status])

    artifact_type = str(((candidate.get("artifact") or {}).get("artifact_type")) or "").lower()
    if artifact_type in weights:
        score += float(weights[artifact_type])
        breakdown[f"artifact:{artifact_type}"] = float(weights[artifact_type])

    validation = candidate.get("validation") or {}
    errors = validation.get("errors") if isinstance(validation, dict) else []
    error_count = len(errors) if isinstance(errors, list) else 0
    if error_count:
        delta = float(weights.get("validation_error", -1.0)) * float(error_count)
        score += delta
        breakdown["validation:error_count"] = delta

    confidence = _read_confidence(candidate)
    if confidence is not None:
        conf_score = float(weights.get("confidence", 0.0)) * float(confidence)
        score += conf_score
        breakdown["confidence"] = conf_score

    return score, breakdown


def _read_confidence(candidate):
    artifact = candidate.get("artifact") or {}
    body = artifact.get("body") if isinstance(artifact, dict) else {}
    if not isinstance(body, dict):
        return None
    raw = body.get("confidence", artifact.get("confidence"))
    try:
        value = float(raw)
    except (TypeError, ValueError):
        return None
    if value < 0:
        return 0.0
    if value > 1:
        return 1.0
    return value


def _is_eligible(candidate, policy_obj):
    if not policy_obj.get("require_validated", True):
        return True
    status = str(candidate.get("status") or "").lower()
    return status in {"validated", "injected"}


def _merge_policy(policy):
    merged = deepcopy(DEFAULT_POLICY)
    if not isinstance(policy, dict):
        return merged
    if isinstance(policy.get("weights"), dict):
        for key, value in policy["weights"].items():
            try:
                merged["weights"][str(key)] = float(value)
            except (TypeError, ValueError):
                continue
    if "require_validated" in policy:
        merged["require_validated"] = bool(policy.get("require_validated"))
    return merged



# hiveos/session_manager.py

import uuid
from threading import RLock
from hiveos.core import Core
import time

class SessionManager:
    def __init__(self):
        self._lock = RLock()
        self.sessions = {}
        self.active_sessions = {}  # 🔄 NEW: {user_id: session_id}
        self.active_session_id = None
        self._label_counter = 1
        self.gc_interval = 30  # seconds between GC sweeps
        self.start_gc_loop()



    def start_gc_loop(self):
        from threading import Thread

        def gc_loop():
            while True:
                time.sleep(self.gc_interval)
                with self._lock:
                    sessions_snapshot = list(self.sessions.items())
                for sid, session in sessions_snapshot:
                    core = session.get("core")
                    if core and self.should_cleanup(core):
                        print(f"[GC] Cleaning up expired session {sid}")
                        try:
                            core.teardown()
                        except Exception as e:
                            print(f"[GC] Error tearing down session {sid}: {e}")
                        with self._lock:
                            self.sessions.pop(sid, None)
                            if sid == self.active_session_id:
                                self.active_session_id = next(iter(self.sessions), None)

        Thread(target=gc_loop, daemon=True).start()


    def create_session(self, label=None, user_id=None):
        session_id = str(uuid.uuid4())
        with self._lock:
            if not label:
                label = f"Session {self._label_counter}"
                self._label_counter += 1
        core = Core(session_id=session_id)
        core.start_background_tick()
        with self._lock:
            self.sessions[session_id] = {
                "core": core,
                "label": label,
                "user_id": user_id
            }
            if user_id:
                self.active_sessions[user_id] = session_id

        print(f"[SessionManager] Created session {session_id}")
        return session_id

    def get_session(self, session_id=None, user_id=None):
        with self._lock:
            if session_id is None and user_id:
                session_id = self.active_sessions.get(user_id)
            return self.sessions.get(session_id, {}).get("core")

    def get_session_record(self, session_id):
        with self._lock:
            session = self.sessions.get(session_id)
            return dict(session) if session else None


    def get_all_sessions(self):
        with self._lock:
            sessions_snapshot = list(self.sessions.items())
        return [
            {
                "id": sid,
                "label": s["label"],
                "user_id": s.get("user_id")
            }
            for sid, s in sessions_snapshot
        ]


    def delete_session(self, session_id):
        with self._lock:
            session = self.sessions.get(session_id)
        if session:
            core = session["core"]
            core.teardown()
            with self._lock:
                self.sessions.pop(session_id, None)
            print(f"[SessionManager] Deleted session {session_id}")
            with self._lock:
                if session_id == self.active_session_id:
                    self.active_session_id = next(iter(self.sessions), None)
            return True
        return False

    def set_active_session(self, session_id, user_id):
        with self._lock:
            if session_id in self.sessions and self.get_session_user_id(session_id) == user_id:
                self.active_sessions[user_id] = session_id
                print(f"[SessionManager] Switched {user_id} to session {session_id}")
                return True
            return False


    def get_user_sessions(self, user_id):
        with self._lock:
            sessions_snapshot = list(self.sessions.items())
        return [
            {"id": sid, "label": s["label"], "user_id": s.get("user_id")}
            for sid, s in sessions_snapshot
            if s.get("user_id") == user_id
        ]

    def get_session_user_id(self, session_id):
        with self._lock:
            return self.sessions.get(session_id, {}).get("user_id")

    def should_cleanup(self, core):
        MAX_SESSION_AGE = 3600  # seconds (30 minutes)
        start = getattr(core, "start_time", 0)
        age = time.time() - start
        return age > MAX_SESSION_AGE




# Global singleton instance
session_manager = SessionManager()

# Create an initial default session immediately
#session_manager.create_session()



import unittest

from hiveos.security.authorization import authorize_action


class AuthorizationPolicyTests(unittest.TestCase):
    def test_unregistered_action_denied(self):
        decision = authorize_action("unknown.action", ["runtime.write"], enforce=True)
        self.assertFalse(decision.allowed)
        self.assertEqual("policy_action_unregistered", decision.reason)

    def test_enforcement_disabled_allows_without_caps(self):
        decision = authorize_action("runtime.inject_yaml", [], enforce=False)
        self.assertTrue(decision.allowed)
        self.assertEqual("capability_enforcement_disabled", decision.reason)

    def test_enforcement_enabled_denies_without_caps(self):
        decision = authorize_action("runtime.inject_yaml", [], enforce=True)
        self.assertFalse(decision.allowed)
        self.assertEqual("capability_denied", decision.reason)

    def test_exact_capability_match(self):
        decision = authorize_action("runtime.inject_yaml", ["runtime.write"], enforce=True)
        self.assertTrue(decision.allowed)

    def test_wildcard_capability_match(self):
        decision = authorize_action("control.pause_core", ["runtime.*"], enforce=True)
        self.assertTrue(decision.allowed)

    def test_admin_capability_match(self):
        decision = authorize_action("proposals.inject", ["admin"], enforce=True)
        self.assertTrue(decision.allowed)

    def test_terminal_exec_capability_match(self):
        decision = authorize_action("terminal.exec", ["terminal.exec"], enforce=True)
        self.assertTrue(decision.allowed)

    def test_intent_preview_capability_match(self):
        decision = authorize_action("intent.preview", ["intelligence.write"], enforce=True)
        self.assertTrue(decision.allowed)

    def test_proposal_archive_capability_match(self):
        decision = authorize_action("proposals.archive", ["intelligence.write"], enforce=True)
        self.assertTrue(decision.allowed)


if __name__ == "__main__":
    unittest.main()



from dataclasses import dataclass


@dataclass(frozen=True)
class AuthorizationDecision:
    allowed: bool
    reason: str
    required_capabilities: tuple[str, ...]
    granted_capabilities: tuple[str, ...]


ACTION_POLICY = {
    "runtime.inject_yaml": ("runtime.write",),
    "agents.inject": ("runtime.write", "hardware.write", "software.write"),
    "agents.inject_pending": ("runtime.write",),
    "agents.remove": ("runtime.write",),
    "agents.input": ("runtime.write",),
    "tasks.reinject": ("runtime.write",),
    "tasks.delete": ("runtime.write",),
    "proposals.validate": ("intelligence.write",),
    "proposals.inject": ("intelligence.write",),
    "proposals.archive": ("intelligence.write",),
    "control.freeze_zone": ("runtime.control",),
    "control.unfreeze_zone": ("runtime.control",),
    "control.kill_zone": ("runtime.control",),
    "control.pause_core": ("runtime.control",),
    "control.resume_core": ("runtime.control",),
    "control.shutdown_core": ("runtime.control",),
    "control.set_tickrate": ("runtime.control",),
    "control.step_tick": ("runtime.control",),
    "sessions.create": ("runtime.session",),
    "sessions.switch": ("runtime.session",),
    "sessions.delete": ("runtime.session",),
    "terminal.open": ("terminal.stream",),
    "terminal.close": ("terminal.stream",),
    "terminal.clear": ("terminal.stream",),
    "terminal.transform": ("terminal.stream",),
    "terminal.exec": ("terminal.exec",),
    "intent.preview": ("intelligence.write",),
    "intent.apply": ("intelligence.write",),
}


def parse_capabilities(raw):
    if raw is None:
        return []
    if isinstance(raw, list):
        return [str(item).strip() for item in raw if str(item).strip()]
    return [item.strip() for item in str(raw).split(",") if item.strip()]


def authorize_action(action, granted_capabilities, enforce=False):
    required = ACTION_POLICY.get(action)
    granted = tuple(parse_capabilities(granted_capabilities))
    if required is None:
        return AuthorizationDecision(
            allowed=False,
            reason="policy_action_unregistered",
            required_capabilities=(),
            granted_capabilities=granted,
        )

    if not enforce and not granted:
        return AuthorizationDecision(
            allowed=True,
            reason="capability_enforcement_disabled",
            required_capabilities=tuple(required),
            granted_capabilities=granted,
        )

    if _has_any_capability(granted, required):
        return AuthorizationDecision(
            allowed=True,
            reason="authorized",
            required_capabilities=tuple(required),
            granted_capabilities=granted,
        )
    return AuthorizationDecision(
        allowed=False,
        reason="capability_denied",
        required_capabilities=tuple(required),
        granted_capabilities=granted,
    )


def _has_any_capability(granted, required):
    return any(_matches_required(grant, req) for grant in granted for req in required)


def _matches_required(granted_capability, required_capability):
    granted = str(granted_capability).strip()
    required = str(required_capability).strip()
    if granted in {"*", "admin"}:
        return True
    if granted == required:
        return True
    if granted.endswith(".*"):
        prefix = granted[:-2]
        return required == prefix or required.startswith(prefix + ".")
    return False



# hiveos/chunk.py
from hiveos.db import get_db

class Chunk:
    @staticmethod
    def create(chunk_id, task_id, required_capability='generic', zone_id=None, payload=None, depends_on=None, ttl=1, timing="ticks", conditional=None, session_id=None):
        with get_db() as conn:
            conn.execute(
                "INSERT INTO task_chunks (chunk_id, task_id, status, assigned_agent, required_capability, zone_id, retry_count, payload, depends_on, ttl, timing, conditional, session_id) "
                "VALUES (?, ?, 'pending', NULL, ?, ?, 0, ?, ?, ?, ?, ?, ?)",
                (chunk_id, task_id, required_capability, zone_id, payload, depends_on, ttl, timing, conditional, session_id)
            )

    @staticmethod
    def get_by_id(chunk_id):
        with get_db() as conn:
            row = conn.execute(
                "SELECT * FROM task_chunks WHERE chunk_id = ?",
                (chunk_id,)
            ).fetchone()
            if row:
                return row
            return None
